from __future__ import annotations


class JijBenchmarkError(Exception):
    """
    Exception for JijBenchmark Errors related to JijBenchmark inherit from this.
    Exception class.
    """


class ConcurrentFailedError(JijBenchmarkError):
    pass


class SolverFailedError(JijBenchmarkError):
    pass


class UserFunctionFailedError(JijBenchmarkError):
    pass



from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)

import jijbench.exceptions.exceptions as exceptions

from jijbench.exceptions.exceptions import (
    ConcurrentFailedError,
    SolverFailedError,
)

__all__ = ["exceptions", "ConcurrentFailedError", "SolverFailedError"]



from __future__ import annotations

import datetime
import jijmodeling as jm
import pandas as pd
import typing as tp

from typing_extensions import TypeAlias

if tp.TYPE_CHECKING:
    from jijbench.containers.containers import Artifact, Record, Table
    from jijbench.experiment.experiment import Experiment
    from jijbench.node.base import DataNode

T = tp.TypeVar("T")
T_co = tp.TypeVar("T_co")

# node
DataNodeT = tp.TypeVar("DataNodeT", bound="DataNode")
DataNodeT2 = tp.TypeVar("DataNodeT2", bound="DataNode")
DataNodeT_co = tp.TypeVar("DataNodeT_co", bound="DataNode", covariant=True)
DataNodeT2_co = tp.TypeVar("DataNodeT2_co", bound="DataNode", covariant=True)

# element
DateTypes: TypeAlias = tp.Union[str, datetime.datetime, pd.Timestamp]
NumberTypes: TypeAlias = tp.Union[int, float]

# solver
ModelType: TypeAlias = tp.Tuple[jm.Problem, jm.PH_VALUES_INTERFACE]


# mapping
MappingT = tp.TypeVar("MappingT", "Artifact", "Experiment", "Record", "Table")
MappingTypes: TypeAlias = tp.Union["Artifact", "Experiment", "Record", "Table"]
MappingListTypes: TypeAlias = tp.Union[
    tp.List["Artifact"], tp.List["Experiment"], tp.List["Record"], tp.List["Table"]
]
ArtifactDataType: TypeAlias = tp.Dict[tp.Hashable, tp.Dict[tp.Hashable, "DataNode"]]
ExperimentDataType: TypeAlias = tp.Tuple["Artifact", "Table"]



import datetime
import numpy as np
import pandas as pd
import pytest
import jijbench as jb


def test_construct_array():
    array = jb.Array(np.array([1, 2]), name="sample")

    assert array.data[0] == 1


def test_construct_date():
    date = jb.Date()

    assert isinstance(date.data, pd.Timestamp)

    date = jb.Date("2023-01-01")

    assert date.data == pd.Timestamp(2023, 1, 1)

    date = jb.Date(datetime.datetime(2023, 1, 1))

    assert date.data == pd.Timestamp(2023, 1, 1)


def test_construct_number():
    x = jb.Number(1, "x")

    assert x.data == 1

    x = jb.Number(1.0, "x")

    assert x.data == 1.0


def test_construct_string():
    s = jb.String("a", "s")

    assert s.data == "a"


def test_construct_callable():
    def f():
        pass

    c = jb.Callable(f, "f")

    assert c.data == f


def test_array_invalid_data():
    with pytest.raises(TypeError):
        jb.Array([1, 2], name="sample")


def test_date_invalid_data():
    with pytest.raises(TypeError):
        jb.Date(1)

    with pytest.raises(ValueError):
        jb.Date("a")


def test_number_invalid_data():
    with pytest.raises(TypeError):
        jb.Number("1", "x")


def test_string_invalid_data():
    with pytest.raises(TypeError):
        jb.String(1, "x")


def test_callable_invalid_data():
    with pytest.raises(TypeError):
        jb.Callable(1, "f")


def test_array_min():
    array = jb.Array(np.arange(5), name="sample")
    res = array.min()

    assert res.data == 0.0
    assert isinstance(res.operator, jb.functions.Min)
    assert array in res.operator.inputs


def test_array_max():
    array = jb.Array(np.arange(5), name="sample")
    res = array.max()

    assert res.data == 4.0
    assert isinstance(res.operator, jb.functions.Max)
    assert array in res.operator.inputs


def test_array_mean():
    array = jb.Array(np.array([1.0, 1.0]), name="sample")
    res = array.mean()

    assert res.data == 1.0
    assert isinstance(res.operator, jb.functions.Mean)
    assert array in res.operator.inputs


def test_array_std():
    array = jb.Array(np.array([1.0, 1.0]), name="sample")
    res = array.std()

    assert res.data == 0.0
    assert isinstance(res.operator, jb.functions.Std)
    assert array in res.operator.inputs



import os, shutil

import jijmodeling as jm
import jijzept as jz
import pytest

import jijbench as jb
from unittest.mock import MagicMock


@pytest.fixture(scope="function", autouse=True)
def pre_post_process():
    # preprocess
    yield
    # postprocess
    norm_path = os.path.normcase("./.jb_results")
    if os.path.exists(norm_path):
        shutil.rmtree(norm_path)


def test_simple_benchmark():
    def func(x):
        return x

    bench = jb.Benchmark({"x": [1, 2]}, solver=func, name="test")

    res = bench(autosave=True)
    columns = res.table.columns

    assert isinstance(res, jb.Experiment)
    assert "func_return[0]" in columns

    op1 = res.operator
    assert op1 is not None
    assert isinstance(op1.inputs[0], jb.Experiment)
    assert isinstance(op1.inputs[1], jb.Experiment)
    t1 = op1.inputs[0].table
    t2 = op1.inputs[1].table

    assert t1.iloc[0, 1] == 1
    assert t2.iloc[0, 1] == 2


def test_benchmark_for_jijzept_sampler(
    sample_model: MagicMock,
    sa_sampler: jz.JijSASampler,
    knapsack_problem: jm.Problem,
    knapsack_instance_data: jm.PH_VALUES_INTERFACE,
):
    bench = jb.construct_benchmark_for(
        sa_sampler,
        [(knapsack_problem, knapsack_instance_data)],
        {"num_reads": [1, 2]},
    )
    res = bench(autosave=False)

    assert sample_model.call_count == 2
    assert len(sample_model.call_args_list) == 2
    sample_model.assert_called_with(
        model=knapsack_problem, feed_dict=knapsack_instance_data
    )

    table = res.table.reset_index()
    assert table.loc[0, "num_samples"] == 10
    assert table.loc[0, "num_feasible"] == 7


def test_benchmark_for_jijzept_sampler_with_multi_models(
    sample_model: MagicMock,
    sa_sampler: jz.JijSASampler,
    knapsack_problem: jm.Problem,
    knapsack_instance_data: jm.PH_VALUES_INTERFACE,
    tsp_problem: jm.Problem,
    tsp_instance_data: jm.PH_VALUES_INTERFACE,
):
    models = [
        (knapsack_problem, knapsack_instance_data),
        (tsp_problem, tsp_instance_data),
    ]
    bench = jb.construct_benchmark_for(
        sa_sampler,
        models,
        {
            "search": [True, False],
            "num_search": [5],
        },
    )
    res = bench(autosave=False)

    assert sample_model.call_count == 4
    assert len(sample_model.call_args_list) == 4

    sample_model.assert_any_call(
        model=knapsack_problem,
        feed_dict=knapsack_instance_data,
        search=True,
        num_search=5,
    )
    sample_model.assert_any_call(
        model=tsp_problem,
        feed_dict=tsp_instance_data,
        search=False,
        num_search=5,
    )

    table = res.table.reset_index()
    assert table.loc[0, "num_samples"] == 10
    assert table.loc[0, "num_feasible"] == 7


def test_benchmark_for_jijzept_sampler_using_params(
    onehot_problem: jm.Problem, jm_sampleset: jm.SampleSet
):
    def f(problem, instance_data, **kwargs) -> jm.SampleSet:
        if not isinstance(problem, jm.Problem):
            raise TypeError
        if not isinstance(instance_data, dict):
            raise TypeError
        return jm_sampleset

    instance_data = {"d": [1 for _ in range(10)]}
    instance_data["d"][0] = -1

    bench = jb.Benchmark(
        {
            "num_reads": [1, 2],
            "num_sweeps": [10],
            "problem": [onehot_problem],
            "instance_data": [instance_data],
        },
        solver=f,
    )
    res = bench(autosave=False)

    # assert res.table["problem_name"][0] == "problem"
    # assert res.table["instance_data_name"][0] == "Unnamed[0]"


def test_apply_benchmark():
    def func(x):
        return x

    bench = jb.Benchmark(
        {"x": [1, 2]},
        solver=func,
    )

    experiment = jb.Experiment(name=jb.ID().data)
    res = experiment.apply(bench)
    columns = res.table.columns

    assert isinstance(res, jb.Experiment)
    assert "func_return[0]" in columns

    op1 = res.operator
    # ic()
    # ic(op1.inputs)
    assert op1 is not None
    assert isinstance(op1, jb.Benchmark)
    assert isinstance(op1.inputs[0], jb.Experiment)
    assert len(op1.inputs) == 1
    assert op1.inputs[0].table.empty


def test_benchmark_params_table():
    def func(x):
        return x

    bench = jb.Benchmark(
        {"x": [1, 2]},
        solver=func,
    )

    res = bench()


def test_benchmark_with_multi_return_solver():
    def func():
        return "a", 1

    bench = jb.Benchmark({"num_reads": [1, 2], "num_sweeps": [10]}, solver=func)
    res = bench()

    assert len(res.table) == 2
    assert res.table["func_return[0]"][0] == "a"
    assert res.table["func_return[1]"][0] == 1.0


# def test_benchmark_with_custom_solver_by_sync_False():
#     def func():
#         return "a", 1
#
#     bench = jb.Benchmark({"num_reads": [1, 2], "num_sweeps": [10]}, solver=func)
#     with pytest.raises(ConcurrentFailedError):
#         bench.run(sync=False)


def test_benchmark_with_callable_args():
    def f(x):
        return x**2

    def rap_solver(x, f):
        return f(x)

    bench = jb.Benchmark(
        {
            "x": [1, 2, 3],
            "f": [f],
        },
        solver=rap_solver,
    )

    res = bench()

    # assert sample_model.__name__ in columns
    # assert isinstance(res.table[sample_model.__name__][0], str)


# def test_benchmark_with_multisolver():
#     def func1(x):
#         return 2 * x
#
#     def func2(x):
#         return 3 * x
#
#     bench = jb.Benchmark(params={"x": [1, 2, 3]}, solver=[func1, func2])
#     bench.run()
#
#     columns = bench.table.columns
#
#     assert "solver" in columns
#     assert "func1" in bench.table["solver"].values
#     assert "func2" in bench.table["solver"].values
#
#
# def test_load():
#     def func1(x):
#         return 2 * x
#
#     bench = jb.Benchmark(params={"x": [1, 2, 3]}, solver=func1, benchmark_id="test")
#     bench.run()
#
#     del bench
#
#     bench = jb.load(benchmark_id="test")
#
#     assert "func1" in bench.table["solver"].values
#
#
# def test_save():
#     def func1(x):
#         return 2 * x
#
#     import pathlib
#
#     save_dir = str(pathlib.PurePath(__file__).parent / ".my_result")
#
#     bench = jb.Benchmark(
#         params={"x": [1, 2, 3]}, solver=func1, benchmark_id="test", save_dir=save_dir
#     )
#     bench.run()
#
#     shutil.rmtree(save_dir)
#
#
# def test_benchmark_for_custom_solver_return_jm_sampleset():
#     def func():
#         jm_sampleset = jm.SampleSet.from_serializable(
#             {
#                 "record": {
#                     "solution": {
#                         "x": [
#                             (([0, 1], [0, 1]), [1, 1], (2, 2)),
#                             (([], []), [], (2, 2)),
#                         ]
#                     },
#                     "num_occurrences": [1, 1],
#                 },
#                 "evaluation": {
#                     "energy": [
#                         -3.8499999046325684,
#                         0.0,
#                     ],
#                     "objective": [3.0, 0.0],
#                     "constraint_violations": {},
#                     "penalty": None,
#                 },
#                 "measuring_time": {
#                     "solve": None,
#                     "system": None,
#                     "total": None,
#                 },
#             }
#         )
#         jm_sampleset.measuring_time.solve.solve = None
#         jm_sampleset.measuring_time.system.system = None
#         jm_sampleset.measuring_time.total = None
#         return jm_sampleset
#
#     bench = jb.Benchmark(params={"dummy": [1]}, solver=func)
#     bench.run()
#
#
# def test_benchmark_for_custom_solver_failed():
#     def custom_solver_failed():
#         raise Exception("solver is failed.")
#
#     bench = jb.Benchmark(params={"dummy": [1]}, solver=custom_solver_failed)
#     with pytest.raises(SolverFailedError):
#         bench.run()
#
#
# def test_benchmark_for_num_feasible():
#     bench = jb.Benchmark(
#         {
#             "N": [10, 200],
#             "sample_model": [sample_model],
#         },
#         solver=sample_model,
#     )
#     bench.run()
#     assert (bench.table["num_feasible"].values == 7).all()
#
#
# def test_benchmark_for_change_solver_return_name():
#     def solver():
#         return 1
#
#     bench = jb.Benchmark(
#         {
#             "N": [10, 200],
#             "sample_model": [sample_model],
#         },
#         solver=solver,
#         solver_return_name={"solver": ["return_1"]},
#     )
#     bench.run()
#     assert "return_1" in bench.table.columns



from __future__ import annotations

from enum import Enum, auto
import matplotlib.pyplot as plt
import networkx as nx
import numpy.typing as npt
from typing import cast

from jijbench.visualization.figure.interface import Figure

# TODO: node_posなどの型アノテーションを行う


class GraphType(Enum):
    UNDIRECTED = auto()
    DIRECTED = auto()


graph_modules = {
    GraphType.UNDIRECTED: nx.Graph,
    GraphType.DIRECTED: nx.DiGraph,
}


class Graph(Figure):
    """Visualize graph.

    You can also instantiate from edge_list, distance matrix by class method.
    Attributes:
        G (nx.Graph): the networkX Graph instance.
        fig_ax (tuple[matplotlib.figure.Figure, matplotlib.axes.Subplot]): Figure and Axes of matplotlib. Available after show method is called.
    Example:
        The code below visualizes an undirected graph given by edge list.
        The style of the graph (e.g. color) can be changed by arguments of the show method.
        ```python
        >>> from jijbench.figure.graph import Graph, GraphType

        >>> graph = Graph.from_edge_list([[1, 2], [2, 3]], GraphType.UNDIRECTED)
        >>> graph.show(edge_color=["black", "red"])
        ```
    """

    def __init__(
        self,
        G: nx.Graph,
    ):
        self.G = G

    def show(
        self,
        figsize: tuple[int | float] | None = None,
        title: str | None = None,
        node_pos: dict | None = None,
        node_color: (str | list[str]) | None = None,
        edge_color: (str | list[str]) | None = None,
        node_labels: dict | None = None,
        edge_labels: dict | None = None,
    ):
        """Visualize graph.

        The arguments of the show method are passed to the plot of matplotlib, networkx.

        Args:
            figsize (tuple[int | float] | None): the size of figure. The default uses matplotlib's default value.
            title (str | None): the title of figure. Defaults to "graph".
            node_pos (dict | None): dict where the key is node and the value is position (np,array([x, y])). The default uses the return of networkx.spring_layout(self.G, seed=1).
            node_color ((str | list[str]) | None): string or list of node color. Defaults to "#1f78b4".
            edge_color ((str | list[str]) | None): string or list of edge color. Defaults to "k".
            node_labels (dict | None): dict where the key is node and the value is label. The default is {node: str(node) for node in self.G.nodes}.
            edge_labels (dict | None): dict where the key is node and the value is label. The default is {}, or weights for weighted graphs.
        """
        G = self.G

        if title is None:
            title = "graph"

        if node_pos is None:
            node_pos = nx.spring_layout(G, seed=1)

        if node_color is None:
            node_color = "#1f78b4"

        if edge_color is None:
            edge_color = "k"

        if node_labels is None:
            node_labels = {node: str(node) for node in G.nodes}

        if edge_labels is None:
            # If G is not weighted graph, edge_labels will be {}.
            edge_labels = nx.get_edge_attributes(G, "weight")

        fig, ax = plt.subplots(figsize=figsize)
        fig.suptitle(title)

        # execute typing.cast to the avoid type checking error caused by the networkx interface.
        node_color = cast("str", node_color)
        nx.draw_networkx_nodes(
            G=G,
            pos=node_pos,
            node_color=node_color,
            ax=ax,
        )

        # execute typing.cast to the avoid type checking error caused by the networkx interface.
        edge_color = cast("str", edge_color)
        nx.draw_networkx_edges(
            G=G,
            pos=node_pos,
            edge_color=edge_color,
            ax=ax,
        )
        nx.draw_networkx_labels(
            G=G,
            pos=node_pos,
            labels=node_labels,
            ax=ax,
        )
        nx.draw_networkx_edge_labels(
            G=G,
            pos=node_pos,
            edge_labels=edge_labels,
            ax=ax,
        )

        self._fig_ax = (fig, ax)

    @property
    def fig_ax(self):
        if self._fig_ax is None:
            raise AttributeError(
                "fig_ax attribute is available after show method is called."
            )
        else:
            return self._fig_ax

    @classmethod
    def from_edge_list(
        cls,
        edge_list: list[list[int]],
        graphtype: GraphType,
    ):
        """To Graph instance from edge list.

        Args:
            edge_list (list[list[int]]): list of edges.
            graphtype (GraphType): GraphType instance of jijbench.
        Example:
            ```python
            >>> from jijbench.figure.graph import Graph, GraphType

            >>> edge_list = [[1, 2], [2, 3], [2, 1]]
            >>> graph = Graph.from_edge_list(edge_list, GraphType.DIRECTED)
            ```
        """
        G = graph_modules[graphtype]()
        node_list = sorted(set([node for edge in edge_list for node in edge]))
        G.add_nodes_from(node_list)
        G.add_edges_from(edge_list)
        return cls(G)

    @classmethod
    def from_distance_matrix(
        cls,
        distance_matrix: list[list[int | float]] | npt.NDArray,
        graphtype: GraphType,
    ):
        """To Graph instance from distance matrix.

        Args:
            distance_matrix (list[list[int | float]] | npt.NDArray): distance matrix. No self-loop is added, so the diagonal can be any number.
            graphtype (GraphType): GraphType instance of jijbench.
        Example:
            ```python
            >>> from jijbench.figure.graph import Graph, GraphType

            >>> distance_matrix = [[-1, 2], [2, -1]]
            >>> graph = Graph.from_distance_matrix(distance_matrix, GraphType.UNDIRECTED)
            ```
        """
        G = graph_modules[graphtype]()
        node_list = [i for i, _ in enumerate(distance_matrix)]
        edge_list = [
            (i, j, w)
            for i, ws in enumerate(distance_matrix)
            for j, w in enumerate(ws)
            if i != j
        ]
        G.add_nodes_from(node_list)
        G.add_weighted_edges_from(edge_list)
        return cls(G)



from __future__ import annotations

import os, shutil


import jijmodeling as jm
import pandas as pd
import pytest

import jijbench as jb
from unittest.mock import MagicMock


@pytest.fixture(scope="function", autouse=True)
def pre_post_process():
    # preprocess
    yield
    # postprocess
    norm_path = os.path.normcase("./.jb_results")
    if os.path.exists(norm_path):
        shutil.rmtree(norm_path)


def test_simple_experiment():
    e = jb.Experiment(name="simple_experiment")

    def func(i):
        return i**2

    for i in range(3):
        solver = jb.Solver(func)
        record = solver([jb.Parameter(i, "i")])
        e.append(record)
    e.save()


def test_simple_experiment_with_context_manager():
    e = jb.Experiment(name="simple_experiment_with_context_manager", autosave=True)

    def func(i):
        return i**2

    for i in range(3):
        with e:
            solver = jb.Solver(func)
            record = solver([jb.Parameter(i, "i")])
            e.append(record)


def test_construct_experiment():
    e = jb.Experiment(name="test")

    a = jb.Artifact({"x": {"0": jb.Number(1, "value")}})
    t = jb.Table(pd.DataFrame([[jb.Number(1, "value")]]))
    e.data = (a, t)

    a = e.artifact
    a.update({"y": 2})

    t = e.table
    t["x"] = [1]


def test_jijmodeling(
    sample_model: MagicMock,
    knapsack_problem: jm.Problem,
    knapsack_instance_data: jm.PH_VALUES_INTERFACE,
):
    experiment = jb.Experiment(autosave=False)

    with experiment:
        solver = jb.Solver(sample_model)
        x1 = jb.Parameter(knapsack_problem, name="model")
        x2 = jb.Parameter(knapsack_instance_data, name="feed_dict")
        record = solver([x1, x2])
        record.name = jb.ID().data
        experiment.append(record)

    droped_table = experiment.table.dropna(axis="columns")

    cols = droped_table.columns
    assert "energy" in cols
    assert "num_feasible" in cols

    assert sample_model.call_count == 1
    sample_model.assert_called_with(
        model=knapsack_problem, feed_dict=knapsack_instance_data
    )



import os, shutil
import jijmodeling as jm

import numpy as np
import pandas as pd
import pytest

import jijbench as jb
from jijbench.exceptions.exceptions import UserFunctionFailedError
from jijbench.visualization.metrics.plot import MetricsPlot


@pytest.fixture(scope="function", autouse=True)
def pre_post_process():
    # preprocess
    yield
    # postprocess
    norm_path = os.path.normcase("./.jb_results")
    if os.path.exists(norm_path):
        shutil.rmtree(norm_path)


def solve():
    d = jm.Placeholder("d", dim=2)
    x = jm.Binary("x", shape=d.shape)
    i = jm.Element("i", d.shape[0])
    j = jm.Element("j", d.shape[1])
    problem = jm.Problem("simple_problem")
    problem += jm.Sum([i, j], d[i, j] * x[i, j])
    problem += jm.Constraint("onehot1", x[i, :] == 1, forall=i)
    problem += jm.Constraint("onehot2", x[:, j] == 1, forall=j)
    jm_sampleset_dict = {
        "record": {
            "solution": {
                "x": [
                    (([0, 1], [0, 1]), [1, 1], (2, 2)),
                    (([0, 1], [1, 0]), [1, 1], (2, 2)),
                    (([], []), [], (2, 2)),
                    (([0, 1], [0, 0]), [1, 1], (2, 2)),
                ]
            },
            "num_occurrences": [4, 3, 2, 1],
        },
        "evaluation": {
            "energy": [3.0, 24.0, 0.0, 20.0],
            "objective": [3.0, 24.0, 0.0, 17.0],
            "constraint_violations": {
                "onehot1": [0.0, 0.0, 2.0, 0.0],
                "onehot2": [0.0, 0.0, 2.0, 2.0],
            },
            "penalty": {},
        },
        "measuring_time": {"solve": None, "system": None, "total": None},
    }
    jm_sampleset = jm.SampleSet.from_serializable(jm_sampleset_dict)
    solving_time = jm.SolvingTime(
        **{"preprocess": 1.0, "solve": 1.0, "postprocess": 1.0}
    )
    jm_sampleset.measuring_time.solve = solving_time
    return jm_sampleset


def solve_no_constraint():
    d = jm.Placeholder("d", dim=2)
    x = jm.Binary("x", shape=d.shape)
    i = jm.Element("i", d.shape[0])
    j = jm.Element("j", d.shape[1])
    problem = jm.Problem("simple_problem")
    problem += jm.Sum([i, j], d[i, j] * x[i, j])
    problem += jm.Constraint("onehot1", x[i, :] == 1, forall=i)
    problem += jm.Constraint("onehot2", x[:, j] == 1, forall=j)
    jm_sampleset_dict = {
        "record": {
            "solution": {
                "x": [
                    (([0, 1], [0, 1]), [1, 1], (2, 2)),
                    (([0, 1], [1, 0]), [1, 1], (2, 2)),
                    (([], []), [], (2, 2)),
                    (([0, 1], [0, 0]), [1, 1], (2, 2)),
                ]
            },
            "num_occurrences": [4, 3, 2, 1],
        },
        "evaluation": {
            "energy": [3.0, 24.0, 0.0, 20.0],
            "objective": [3.0, 24.0, 0.0, 17.0],
            "constraint_violations": {},
            "penalty": {},
        },
        "measuring_time": {"solve": None, "system": None, "total": None},
    }
    jm_sampleset = jm.SampleSet.from_serializable(jm_sampleset_dict)
    solving_time = jm.SolvingTime(
        **{"preprocess": 1.0, "solve": 1.0, "postprocess": 1.0}
    )
    jm_sampleset.measuring_time.solve = solving_time
    return jm_sampleset


def solve_no_obj_no_constraint():
    d = jm.Placeholder("d", dim=2)
    x = jm.Binary("x", shape=d.shape)
    i = jm.Element("i", d.shape[0])
    j = jm.Element("j", d.shape[1])
    problem = jm.Problem("simple_problem")
    problem += jm.Sum([i, j], d[i, j] * x[i, j])
    problem += jm.Constraint("onehot1", x[i, :] == 1, forall=i)
    problem += jm.Constraint("onehot2", x[:, j] == 1, forall=j)
    jm_sampleset_dict = {
        "record": {
            "solution": {
                "x": [
                    (([0, 1], [0, 1]), [1, 1], (2, 2)),
                    (([0, 1], [1, 0]), [1, 1], (2, 2)),
                    (([], []), [], (2, 2)),
                    (([0, 1], [0, 0]), [1, 1], (2, 2)),
                ]
            },
            "num_occurrences": [4, 3, 2, 1],
        },
        "evaluation": {
            "energy": [3.0, 24.0, 0.0, 20.0],
            "constraint_violations": {},
            "penalty": {},
        },
        "measuring_time": {"solve": None, "system": None, "total": None},
    }
    jm_sampleset = jm.SampleSet.from_serializable(jm_sampleset_dict)
    solving_time = jm.SolvingTime(
        **{"preprocess": 1.0, "solve": 1.0, "postprocess": 1.0}
    )
    jm_sampleset.measuring_time.solve = solving_time
    return jm_sampleset


def fig_contain_target_data(fig, label, values):
    for data in fig.data[0].dimensions:
        if data["label"] == label and (data["values"] == values).all():
            return True
    return False


def test_metrics_plot_parallelplot_num_feasible(mocker):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    bench = jb.Benchmark(
        params={"multipliers": [{}, {}]},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig = mplot.parallelplot_experiment()

    expect_value_0 = result.table["num_feasible"].values[0]
    expect_value_1 = result.table["num_feasible"].values[1]
    expect_arr = np.array(expect_value_0, expect_value_1)

    assert fig_contain_target_data(fig, "num_feasible", expect_arr)


def test_metrics_plot_parallelplot_samplemean_objective(mocker):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    bench = jb.Benchmark(
        params={"multipliers": [{}, {}]},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig = mplot.parallelplot_experiment()

    def calc_expect_samplemean_objective(result, idx):
        obj = result.table["objective"].values[idx]
        occ = result.table["num_occurrences"].values[idx]
        return np.sum(occ * obj) / np.sum(occ)

    expect_mean_0 = calc_expect_samplemean_objective(result, 0)
    expect_mean_1 = calc_expect_samplemean_objective(result, 1)
    expect_arr = np.array(expect_mean_0, expect_mean_1)

    assert fig_contain_target_data(fig, "samplemean_objective", expect_arr)


def test_metrics_plot_parallelplot_samplemean_violations(mocker):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig = mplot.parallelplot_experiment()

    def calc_result_samplemean_violations(result, idx, constraint_name):
        violation = result.table[f"{constraint_name}_violations"].values[idx]
        occ = result.table["num_occurrences"].values[idx]
        return np.sum(occ * violation) / np.sum(occ)

    expect_mean_onehot1 = calc_result_samplemean_violations(result, 0, "onehot1")
    expect_mean_onehot2 = calc_result_samplemean_violations(result, 0, "onehot2")
    expect_mean_total = expect_mean_onehot1 + expect_mean_onehot2

    assert fig_contain_target_data(
        fig, "samplemean_onehot1_violations", np.array([expect_mean_onehot1])
    )

    assert fig_contain_target_data(
        fig, "samplemean_onehot2_violations", np.array([expect_mean_onehot2])
    )

    assert fig_contain_target_data(
        fig, "samplemean_total_violations", expect_mean_total
    )


def test_metrics_plot_parallelplot_samplemean_violations_no_constraint(mocker):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    bench = jb.Benchmark(
        params={},
        solver=[solve_no_constraint],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig = mplot.parallelplot_experiment()

    # samplemean_total_violationsが含まれていないことを確認する（本テストでは制約がない問題を解く状況を想定しているため）
    def no_samplemean_total_violations(fig):
        for data in fig.data[0].dimensions:
            if data["label"] == "samplemean_total_violations":
                return False
        return True

    assert no_samplemean_total_violations(fig)


def test_metrics_plot_parallelplot_multipliers(mocker):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    params = {
        "multipliers": [{"onehot1": 1, "onehot2": 2}, {"onehot1": 3, "onehot2": 4}]
    }

    bench = jb.Benchmark(
        params=params,
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig = mplot.parallelplot_experiment()

    assert fig_contain_target_data(fig, "onehot1_multiplier", np.array([1, 3]))
    assert fig_contain_target_data(fig, "onehot2_multiplier", np.array([2, 4]))


params = {
    "have_constraint_sampleset": (solve, "samplemean_total_violations"),
    "no_constraint_sampleset": (solve_no_constraint, "samplemean_objective"),
    "no_obj_no_constraint_sampleset": (solve_no_obj_no_constraint, None),
}


# TODO 後で直す
@pytest.mark.parametrize(
    "solver, expect",
    list(params.values()),
    ids=list(params.keys()),
)
def test_metrics_plot_parallelplot_arg_color_column_default(mocker, solver, expect):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    bench = jb.Benchmark(
        params={},
        solver=[solver],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig = mplot.parallelplot_experiment()
    assert fig.layout.coloraxis.colorbar.title.text == expect


def test_metrics_plot_parallelplot_arg_color_column(mocker):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig = mplot.parallelplot_experiment(
        color_column_name="samplemean_onehot1_violations"
    )

    assert fig.layout.coloraxis.colorbar.title.text == "samplemean_onehot1_violations"


def test_metrics_plot_parallelplot_arg_color_midpoint(mocker):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig = mplot.parallelplot_experiment(
        color_midpoint=5,
    )

    assert fig.layout.coloraxis.cmid == 5


def test_metrics_plot_parallelplot_arg_color_midpoint_default(mocker):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    bench = jb.Benchmark(
        params={"multipliers": [{}, {}]},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig = mplot.parallelplot_experiment()

    def calc_mean_total_violaions(x: pd.Series):
        num_occ = x["num_occurrences"]
        onehot1_violations = x["onehot1_violations"]
        onehot2_violations = x["onehot2_violations"]
        total_violations_mean = np.sum(
            num_occ * onehot1_violations + num_occ * onehot2_violations
        ) / np.sum(num_occ)
        return total_violations_mean

    mean_total_violations = result.table.apply(calc_mean_total_violaions, axis=1).mean()

    assert fig.layout.coloraxis.cmid == mean_total_violations


params = {
    "give_title_case": ("title", "title"),
    "default_case": (None, None),
}


@pytest.mark.parametrize(
    "title, expect",
    list(params.values()),
    ids=list(params.keys()),
)
def test_metrics_plot_parallelplot_arg_title(mocker, title, expect):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig = mplot.parallelplot_experiment(
        title=title,
    )

    assert fig.layout.title.text == expect


params = {
    "give_height_case": (1000, 1000),
    "default_case": (None, None),
}


@pytest.mark.parametrize(
    "height, expect",
    list(params.values()),
    ids=list(params.keys()),
)
def test_metrics_plot_parallelplot_arg_height(mocker, height, expect):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig = mplot.parallelplot_experiment(
        height=height,
    )

    assert fig.layout.height == expect


params = {
    "give_width_case": (1000, 1000),
    "default_case": (None, None),
}


@pytest.mark.parametrize(
    "width, expect",
    list(params.values()),
    ids=list(params.keys()),
)
def test_metrics_plot_parallelplot_arg_width(mocker, width, expect):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig = mplot.parallelplot_experiment(
        width=width,
    )

    assert fig.layout.width == expect


params = {
    "give_label_pos_case": ("bottom", "bottom"),
    "default_case": (None, "top"),
}


@pytest.mark.parametrize(
    "axis_label_pos, expect",
    list(params.values()),
    ids=list(params.keys()),
)
def test_metrics_plot_parallelplot_arg_axis_label_pos(mocker, axis_label_pos, expect):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig = mplot.parallelplot_experiment(axis_label_pos=axis_label_pos)
    assert fig.data[0].labelside == expect


params = {
    "give_fontsize_case": (20, 20),
    "default_case": (None, None),
}


def test_metrics_plot_parallelplot_arg_axis_label_pos_invalid_value(mocker):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    with pytest.raises(ValueError) as e:
        mplot.parallelplot_experiment(axis_label_pos="INVALID_VALUE")

    assert (
        str(e.value)
        == "axis_label_pos must be 'top' or 'bottom', but INVALID_VALUE is given."
    )


@pytest.mark.parametrize(
    "fontsize, expect",
    list(params.values()),
    ids=list(params.keys()),
)
def test_metrics_plot_parallelplot_arg_axis_label_fontsize(mocker, fontsize, expect):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig = mplot.parallelplot_experiment(axis_label_fontsize=fontsize)
    assert fig.data[0].labelfont.size == expect


def test_metrics_plot_parallelplot_arg_additional_axes(mocker):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig = mplot.parallelplot_experiment(
        additional_axes=["num_samples"],
    )

    expect_num_samples = result.table["num_samples"].values

    assert fig_contain_target_data(fig, "num_samples", expect_num_samples)


def test_metrics_plot_parallelplot_arg_additional_axes_not_number(mocker):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    with pytest.raises(TypeError):
        mplot.parallelplot_experiment(
            additional_axes=["energy"],
        )


def test_metrics_plot_parallelplot_arg_additional_axes_created_by_function(mocker):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)

    def get_max_num_occurrences(x: pd.Series) -> float:
        return np.max(x["num_occurrences"])

    def calc_samplemean_energy(x: pd.Series) -> float:
        num_occ = x["num_occurrences"]
        array = x["energy"]
        mean = np.sum(num_occ * array) / np.sum(num_occ)
        return mean

    fig = mplot.parallelplot_experiment(
        additional_axes_created_by_function={
            "max_num_occurrences": get_max_num_occurrences,
            "samplemean_energy": calc_samplemean_energy,
        },
    )

    def calc_expect_max_num_occurrences(result, idx):
        num_occ = result.table["num_occurrences"].values[idx]
        return np.max(num_occ)

    def calc_expect_samplemean_energy(result, idx):
        energy = result.table["energy"].values[idx]
        num_occ = result.table["num_occurrences"].values[idx]
        return np.sum(num_occ * energy) / np.sum(num_occ)

    expect_max_num_occurrences = calc_expect_max_num_occurrences(result, 0)
    expect_samplemean_energy = calc_expect_samplemean_energy(result, 0)

    assert fig_contain_target_data(
        fig, "max_num_occurrences", np.array([expect_max_num_occurrences])
    )
    assert fig_contain_target_data(
        fig, "samplemean_energy", np.array([expect_samplemean_energy])
    )


def test_metrics_plot_parallelplot_arg_additional_axes_created_by_function_return_not_number(
    mocker,
):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()

    def get_max_num_occurrences_return_not_number(x: pd.Series) -> str:
        return str(np.max(x["num_occurrences"]))

    mplot = MetricsPlot(result)

    with pytest.raises(TypeError):
        mplot.parallelplot_experiment(
            additional_axes_created_by_function={
                "max_num_occurrences": get_max_num_occurrences_return_not_number
            }
        )


def test_metrics_plot_parallelplot_arg_additional_axes_created_by_function_error_in_func(
    mocker,
):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()

    def error_func(x: pd.Series) -> str:
        raise RuntimeError

    mplot = MetricsPlot(result)
    with pytest.raises(UserFunctionFailedError):
        mplot.parallelplot_experiment(
            additional_axes_created_by_function={"column_name": error_func}
        )


def test_metrics_plot_parallelplot_arg_display_axes_list(mocker):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig = mplot.parallelplot_experiment(
        display_axes_list=[
            "samplemean_onehot2_violations",
            "samplemean_onehot1_violations",
        ]
    )

    assert len(fig.data[0].dimensions) == 2
    assert fig.data[0].dimensions[0].label == "samplemean_onehot2_violations"
    assert fig.data[0].dimensions[1].label == "samplemean_onehot1_violations"


def test_metrics_plot_parallelplot_property_parallelplot_axes_list(mocker):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig = mplot.parallelplot_experiment()

    expect_name_list = [dimension.label for dimension in fig.data[0].dimensions]

    assert mplot.parallelplot_axes_list == expect_name_list


def test_metrics_plot_parallelplot_property_parallelplot_axes_list_before_call_parallelplot():
    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)

    assert mplot.parallelplot_axes_list == []


def test_metrics_plot_parallelplot_arg_rename_map(mocker):
    # Without this mock, the browser will launch and display the visualization results
    mocker.patch("plotly.io.show")
    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig = mplot.parallelplot_experiment(
        rename_map={
            "samplemean_onehot1_violations": "Samplemean Onehot1 Violations",
            "samplemean_onehot2_violations": "Samplemean Onehot2 Violations",
            "samplemean_total_violations": "Samplemean Total Violations",
            "hoge": "Hoge",
        }
    )

    def calc_result_samplemean_violations(result, idx, constraint_name):
        violation = result.table[f"{constraint_name}_violations"].values[idx]
        occ = result.table["num_occurrences"].values[idx]
        return np.sum(occ * violation) / np.sum(occ)

    expect_mean_onehot1 = calc_result_samplemean_violations(result, 0, "onehot1")
    expect_mean_onehot2 = calc_result_samplemean_violations(result, 0, "onehot2")
    expect_mean_total = expect_mean_onehot1 + expect_mean_onehot2

    assert fig_contain_target_data(
        fig, "Samplemean Onehot1 Violations", np.array([expect_mean_onehot1])
    )

    assert fig_contain_target_data(
        fig, "Samplemean Onehot2 Violations", np.array([expect_mean_onehot2])
    )

    assert fig_contain_target_data(
        fig, "Samplemean Total Violations", expect_mean_total
    )



import numpy as np
import pandas as pd
import jijbench as jb


def test_record_factory():
    factory = jb.functions.RecordFactory()

    inputs = [jb.ID("id"), jb.Date(), jb.Array(np.arange(5), "array")]
    record = factory(inputs)

    assert isinstance(record, jb.Record)
    assert record.operator is None
    for node in inputs:
        assert node in factory.inputs


def test_table_factory():
    factory = jb.functions.TableFactory()

    data = [jb.ID("id"), jb.Date(), jb.Array(np.arange(5), "array")]
    r1 = jb.Record(pd.Series(data), "a")
    r2 = jb.Record(pd.Series(data), "b")
    table = factory([r1, r2])

    assert isinstance(table, jb.Table)
    assert table.operator is None
    assert table.data.index[0] == "a"
    assert table.data.index[1] == "b"
    for i, d in enumerate(table.data.loc["a"]):
        assert d == factory.inputs[0].data[i]


def test_artifact_factory():
    factory = jb.functions.ArtifactFactory()

    data = [jb.ID("id"), jb.Date(), jb.Array(np.arange(5), "array")]
    r1 = jb.Record(pd.Series(data), "a")
    r2 = jb.Record(pd.Series(data), "b")
    artifact = factory([r1, r2])

    assert isinstance(artifact, jb.Artifact)
    assert artifact.operator is None
    assert "a" in artifact.data
    assert "b" in artifact.data
    for i, d in artifact.data["a"].items():
        assert d == factory.inputs[0].data[i]


def test_concat_record():
    concat = jb.functions.Concat()

    data = [jb.ID("id"), jb.Date(), jb.Array(np.arange(5), "array")]
    r1 = jb.Record(pd.Series(data), "a")
    r2 = jb.Record(pd.Series(data), "b")
    record = concat([r1, r2])

    assert isinstance(record, jb.Record)
    assert len(record.data) == len(r1.data) + len(r2.data)


def test_concat_table():
    concat = jb.functions.Concat()

    factory = jb.functions.TableFactory()
    data = [jb.ID("id"), jb.Date(), jb.Array(np.arange(5), "array")]
    r1 = jb.Record(pd.Series(data), "a")
    r2 = jb.Record(pd.Series(data), "b")

    t1 = factory([r1])
    t2 = factory([r2])

    table = concat([t1, t2])

    assert isinstance(table, jb.Table)
    assert table.operator is None
    assert table.data.index[0] == "a"
    assert table.data.index[1] == "b"
    for i, d in enumerate(table.data.loc["a"]):
        assert d == factory.inputs[0].data[i]


def test_concat_artifact():
    concat = jb.functions.Concat()

    factory = jb.functions.ArtifactFactory()
    data = [jb.ID("id"), jb.Date(), jb.Array(np.arange(5), "array")]
    r1 = jb.Record(pd.Series(data), "a")
    r2 = jb.Record(pd.Series(data), "b")

    a1 = factory([r1])
    a2 = factory([r2])

    artifact = concat([a1, a2])

    print(artifact.data)
    print(artifact.data.keys())

    assert isinstance(artifact, jb.Artifact)
    assert artifact.operator is None
    assert "a" in artifact.data
    assert "b" in artifact.data
    for i, d in artifact.data["a"].items():
        assert d == factory.inputs[0].data[i]


def test_concat_experiment():
    data1 = [jb.ID("id1"), jb.Date(), jb.Number(1, "num1")]
    data2 = [jb.ID("id2"), jb.Date(), jb.Number(2, "num2")]
    r1 = jb.Record(pd.Series(data1), "a")
    r2 = jb.Record(pd.Series(data2), "b")

    factory = jb.functions.ArtifactFactory()
    a1 = factory([r1])
    a2 = factory([r2])

    factory = jb.functions.TableFactory()
    t1 = factory([r1])
    t2 = factory([r2])

    e1 = jb.Experiment((a1, t1), name="test1")
    e2 = jb.Experiment((a2, t2), name="test2")

    concat = jb.functions.Concat()
    e = concat([e1, e2])
    print()
    print(e.artifact)
    print(e.table)



from __future__ import annotations

from collections import OrderedDict
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt

from jijbench.visualization.figure.interface import Figure


class TimeSeries(Figure):
    """Visualize time series.

    Attributes:
        data (OrderedDict):the dict of time series. the key is label, and the value is tuple of x and y.
        fig_ax (tuple[matplotlib.figure.Figure, matplotlib.axes.Subplot]): Figure and Axes of matplotlib. Available after show method is called.
    Example:
        The code below plots a linear function and a quadratic function.
        The style of the graph (e.g. color) can be changed by arguments of the show method.

        ```python
        >>> import numpy as np
        >>> from jijbench.figure.timeseries import TimeSeries

        >>> x1 = np.arange(-10, 11, 1)
        >>> y1 = x1 + 1

        >>> x2 = np.arange(-10, 11, 1)
        >>> y2 = x2 ** 2

        >>> timeseries = TimeSeries()
        >>> timeseries.add_data("linear", x1, y1)
        >>> timeseries.add_data("quadratic", x2, y2)
        >>> timeseries.show(color_list=["red", "green"])
        ```
    """

    def __init__(
        self,
    ) -> None:
        self.data = OrderedDict()
        self._fig_ax = None

    def add_data(
        self,
        label: str,
        plot_x: list[int | float] | npt.NDArray,
        plot_y: list[int | float] | npt.NDArray,
    ) -> None:
        """Add time series data to data attribute for plot.

        Args:
            label (str): the label of the time series.
            plot_x (list[int | float] | npt.NDArray): the 1D list of horizontal axis value (the list of time).
            plot_y (list[int | float] | npt.NDArray): the 1D list of vertical axis value.
        """
        if isinstance(plot_x, np.ndarray):
            plot_x = plot_x.tolist()
        if isinstance(plot_y, np.ndarray):
            plot_y = plot_y.tolist()

        if len(plot_x) != len(plot_y):
            raise ValueError("plot_x and plot_y must be the same length.")
        self.data.update([(label, (plot_x, plot_y))])

    def show(
        self,
        figsize: tuple[int | float] | None = None,
        title: str | None = None,
        color_list: list | None = None,
        alpha_list: list[float] | None = None,
        linestyle_list: list[str] | None = None,
        marker_list: list[str] | None = None,
        xlabel: str | None = None,
        ylabel: str | None = None,
        xticks: list[int | float] | None = None,
        yticks: list[int | float] | None = None,
    ):
        """Plot time series data which you passed to the add_data method.

        The arguments of the show method are passed to the plot of matplotlib.

        Args:
            figsize (tuple[int | float] | None): the size of figure. The default uses matplotlib's default value.
            title (str | None): the title of figure. Defaults to "time series".
            color_list (list | None): the list of plot line color. The default uses matplotlib's default value.
            alpha_list (list[float] | None): the list of plot line transparency. The default is 1.0 for each plot line.
            linestyle_list (list[str] | None): the list of plot line linestyle. The default is "solid" for each plot line.
            marker_list (list[str] | None): the list of plot line marker. The default is "o" for each plot line.
            xlabel (str | None): the xlabel of figure. Defaults to None.
            ylabel (str | None): the ylabel of figure. Defaults to None.
            xticks (list[int | float] | None): the xticks of figure. The default uses matplotlib's default.
            yticks (list[int | float] | None): the yticks of figure. The default uses matplotlib's default.
        """
        data = self.data

        if len(data) == 0:
            raise RuntimeError(
                "no plot data. Add at least one plot data with add_data method."
            )

        if title is None:
            title = "time series"

        if (color_list is not None) and (len(color_list) != len(data)):
            raise ValueError("color_list and data must be same length.")

        if alpha_list is None:
            alpha_list = [1.0] * len(data)
        elif len(alpha_list) != len(data):
            raise ValueError("alpha_list and data must be same length.")

        if linestyle_list is None:
            linestyle_list = ["solid"] * len(data)
        elif len(linestyle_list) != len(data):
            raise ValueError("linestyle_list and data must be same length.")

        if marker_list is None:
            marker_list = ["o"] * len(data)
        elif len(marker_list) != len(data):
            raise ValueError("marker_list and data must be same length.")

        if xlabel is None:
            xlabel = "time"

        fig, ax = plt.subplots(figsize=figsize)
        fig.suptitle(title)

        for i, (label, plot_data) in enumerate(data.items()):
            x, y = plot_data
            color = None if color_list is None else color_list[i]
            ax.plot(
                x,
                y,
                label=label,
                color=color,
                alpha=alpha_list[i],
                linestyle=linestyle_list[i],
                marker=marker_list[i],
            )

        ax.set_xlabel(xlabel)
        if ylabel is not None:
            ax.set_ylabel(ylabel)
        if xticks is not None:
            ax.set_xticks(xticks)
        if yticks is not None:
            ax.set_yticks(yticks)
        ax.legend()

        self._fig_ax = (fig, ax)

    @property
    def fig_ax(self):
        if self._fig_ax is None:
            raise AttributeError(
                "fig_ax attribute is available after show method is called."
            )
        else:
            return self._fig_ax



from collections import OrderedDict
import matplotlib
from matplotlib import axes, collections, figure, patches
import networkx as nx
import numpy as np
import pytest

from jijbench.visualization.figure.graph import Graph, GraphType
from jijbench.visualization.figure.schedule import Schedule
from jijbench.visualization.figure.timeseries import TimeSeries

# TODO: colorについて、strだけじゃなくてrgb tupleにも対応させる（テストの追記と型アノテーションでいけると想定）
# 参考: rgb tupleについての言及 https://networkx.org/documentation/stable/reference/generated/networkx.drawing.nx_pylab.draw_networkx_nodes.html

params = {
    "list case": ("data", [1, 2], [3, 4]),
    "np.ndarray case": ("data", np.array([1, 2]), np.array([3, 4])),
}


@pytest.mark.parametrize(
    "label, plot_x, plot_y",
    list(params.values()),
    ids=list(params.keys()),
)
def test_timeseries_add_data(label, plot_x, plot_y):
    timeseries = TimeSeries()
    timeseries.add_data(label, plot_x, plot_y)

    assert timeseries.data == OrderedDict([("data", ([1, 2], [3, 4]))])


def test_timeseries_add_data_not_same_length():
    timeseries = TimeSeries()

    with pytest.raises(ValueError):
        timeseries.add_data("data", [1, 2], [3, 4, 5])


def test_timeseries_fig_ax_attribute():
    timeseries = TimeSeries()
    timeseries.add_data("data", [1, 2], [3, 4])
    timeseries.show()
    fig, ax = timeseries.fig_ax

    assert type(fig) == figure.Figure
    assert type(ax) == axes.Subplot


def test_timeseries_fig_ax_attribute_before_show():
    timeseries = TimeSeries()
    timeseries.add_data("data", [1, 2], [3, 4])

    with pytest.raises(AttributeError):
        timeseries.fig_ax


def test_timeseries_show_no_plot_data():
    timeseries = TimeSeries()
    with pytest.raises(RuntimeError):
        timeseries.show()


params = {
    "give argument": ("title", "title"),
    "default": (None, "time series"),
}


@pytest.mark.parametrize(
    "title, expect",
    list(params.values()),
    ids=list(params.keys()),
)
def test_timeseries_show_arg_title(title, expect):
    timeseries = TimeSeries()
    timeseries.add_data("data", [1, 2], [3, 4])
    timeseries.show(title=title)
    fig, ax = timeseries.fig_ax

    success_show_title = False
    for obj in fig.texts:
        actual_title = obj.get_text()
        if actual_title == expect:
            success_show_title = True

    assert success_show_title


def test_timeseries_show_x_and_y():
    x1, y1 = [1, 2], [3, 4]
    x2, y2 = [5, 6], [7, 8]

    timeseries = TimeSeries()
    timeseries.add_data("data1", x1, y1)
    timeseries.add_data("data2", x2, y2)
    timeseries.show()
    fig, ax = timeseries.fig_ax

    assert (ax.get_lines()[0].get_xdata() == np.array(x1)).all()
    assert (ax.get_lines()[0].get_ydata() == np.array(y1)).all()
    assert (ax.get_lines()[1].get_xdata() == np.array(x2)).all()
    assert (ax.get_lines()[1].get_ydata() == np.array(y2)).all()


def test_timeseries_show_arg_figsize():
    figwidth, figheight = 8, 4

    timeseries = TimeSeries()
    timeseries.add_data("data", [1, 2], [3, 4])
    timeseries.show(figsize=tuple([figwidth, figheight]))
    fig, ax = timeseries.fig_ax

    assert fig.get_figwidth() == 8
    assert fig.get_figheight() == 4


def test_timeseries_show_arg_color_list():
    color_list = ["r", "#e41a1c"]

    timeseries = TimeSeries()
    timeseries.add_data("data0", [1, 2], [3, 4])
    timeseries.add_data("data1", [1, 2], [3, 4])
    timeseries.show(color_list=color_list)
    fig, ax = timeseries.fig_ax

    assert ax.get_lines()[0].get_color() == "r"
    assert ax.get_lines()[1].get_color() == "#e41a1c"


def test_timeseries_show_arg_color_list_invalid_length():
    color_list = ["r", "g", "b"]

    timeseries = TimeSeries()
    timeseries.add_data("data0", [1, 2], [3, 4])
    timeseries.add_data("data1", [1, 2], [3, 4])

    with pytest.raises(ValueError):
        timeseries.show(color_list=color_list)


params = {
    "give argument": ([0.5, 0.7], 0.5, 0.7),
    "default": (None, 1.0, 1.0),
}


@pytest.mark.parametrize(
    "alpha_list, expect_alpha_1, expect_alpha_2",
    list(params.values()),
    ids=list(params.keys()),
)
def test_timeseries_show_arg_alpha_list(alpha_list, expect_alpha_1, expect_alpha_2):
    timeseries = TimeSeries()
    timeseries.add_data("data0", [1, 2], [3, 4])
    timeseries.add_data("data1", [1, 2], [3, 4])
    timeseries.show(alpha_list=alpha_list)
    fig, ax = timeseries.fig_ax

    assert ax.get_lines()[0].get_alpha() == expect_alpha_1
    assert ax.get_lines()[1].get_alpha() == expect_alpha_2


def test_timeseries_show_arg_alpha_list_invalid_length():
    alpha_list = [0.1, 0.1, 0.1]

    timeseries = TimeSeries()
    timeseries.add_data("data0", [1, 2], [3, 4])
    timeseries.add_data("data1", [1, 2], [3, 4])

    with pytest.raises(ValueError):
        timeseries.show(alpha_list=alpha_list)


params = {
    "give argument": (["-", "--"], "-", "--"),
    "default": (None, "-", "-"),
}


@pytest.mark.parametrize(
    "linestyle_list, expect_linestyle_1, expect_linestyle_2",
    list(params.values()),
    ids=list(params.keys()),
)
def test_timeseries_show_arg_linestyle_list(
    linestyle_list, expect_linestyle_1, expect_linestyle_2
):
    timeseries = TimeSeries()
    timeseries.add_data("data0", [1, 2], [3, 4])
    timeseries.add_data("data1", [1, 2], [3, 4])
    timeseries.show(linestyle_list=linestyle_list)
    fig, ax = timeseries.fig_ax

    assert ax.get_lines()[0].get_linestyle() == expect_linestyle_1
    assert ax.get_lines()[1].get_linestyle() == expect_linestyle_2


def test_timeseries_show_arg_linestyle_list_invalid_length():
    linestyle_list = ["--", "--", "--"]

    timeseries = TimeSeries()
    timeseries.add_data("data0", [1, 2], [3, 4])
    timeseries.add_data("data1", [1, 2], [3, 4])

    with pytest.raises(ValueError):
        timeseries.show(linestyle_list=linestyle_list)


params = {
    "give argument": (["v", "d"], "v", "d"),
    "default": (None, "o", "o"),
}


@pytest.mark.parametrize(
    "marker_list, expect_marker_1, expect_marker_2",
    list(params.values()),
    ids=list(params.keys()),
)
def test_timeseries_show_arg_marker_list(marker_list, expect_marker_1, expect_marker_2):
    timeseries = TimeSeries()
    timeseries.add_data("data0", [1, 2], [3, 4])
    timeseries.add_data("data1", [1, 2], [3, 4])
    timeseries.show(marker_list=marker_list)
    fig, ax = timeseries.fig_ax

    assert ax.get_lines()[0].get_marker() == expect_marker_1
    assert ax.get_lines()[1].get_marker() == expect_marker_2


def test_timeseries_show_arg_marker_list_invalid_length():
    marker_list = ["v", "v", "v"]

    timeseries = TimeSeries()
    timeseries.add_data("data0", [1, 2], [3, 4])
    timeseries.add_data("data1", [1, 2], [3, 4])

    with pytest.raises(ValueError):
        timeseries.show(marker_list=marker_list)


params = {
    "give argument": ("xlabel", "xlabel"),
    "default": (None, "time"),
}


@pytest.mark.parametrize(
    "xlabel, expect",
    list(params.values()),
    ids=list(params.keys()),
)
def test_timeseries_show_arg_xlabel(xlabel, expect):
    timeseries = TimeSeries()
    timeseries.add_data("data", [1, 2], [3, 4])
    timeseries.show(xlabel=xlabel)
    fig, ax = timeseries.fig_ax

    assert ax.get_xlabel() == expect


def test_timeseries_show_arg_ylabel():
    ylabel = "ylabel"

    timeseries = TimeSeries()
    timeseries.add_data("data", [1, 2], [3, 4])
    timeseries.show(ylabel=ylabel)
    fig, ax = timeseries.fig_ax

    assert ax.get_ylabel() == "ylabel"


def test_timeseries_show_arg_xticks():
    xticks = [1.0, 1.5, 2.0]

    timeseries = TimeSeries()
    timeseries.add_data("data", [1, 2], [3, 4])
    timeseries.show(xticks=xticks)
    fig, ax = timeseries.fig_ax

    assert (ax.get_xticks() == np.array([1.0, 1.5, 2.0])).all()


def test_timeseries_show_arg_yticks():
    yticks = [3.0, 3.5, 4.0]

    timeseries = TimeSeries()
    timeseries.add_data("data", [1, 2], [3, 4])
    timeseries.show(yticks=yticks)
    fig, ax = timeseries.fig_ax

    assert (ax.get_yticks() == np.array([3.0, 3.5, 4.0])).all()


params = {
    "list case": ("data", [1, 2], [3, 4], [5.5, 6.6]),
    "np.ndarray case": (
        "data",
        np.array([1, 2]),
        np.array([3, 4]),
        np.array([5.5, 6.6]),
    ),
}


@pytest.mark.parametrize(
    "task_label, workers, start_times, time_lengths",
    list(params.values()),
    ids=list(params.keys()),
)
def test_schedule_add_data(task_label, workers, start_times, time_lengths):
    schedule = Schedule()
    schedule.add_data(task_label, workers, start_times, time_lengths)

    assert schedule.data == OrderedDict([("data", ([1, 2], [3, 4], [5.5, 6.6]))])


def test_schedule_add_data_attribute_workers():
    schedule = Schedule()
    schedule.add_data("data1", [1, 2], [3, 4], [2, 4])
    schedule.add_data("data2", [2, 3], [7, 8], [9, 10])

    assert schedule.workers == [1, 2, 3]


params = {
    "workers and start_times are different": ("data", [1, 2], [3, 4, 5], [5.5, 6.6]),
    "workers and time_lengths are different": (
        "data",
        np.array([1, 2]),
        np.array([3, 4]),
        np.array([5.5, 6.6, 7.7]),
    ),
}


@pytest.mark.parametrize(
    "task_label, workers, start_times, time_lengths",
    list(params.values()),
    ids=list(params.keys()),
)
def test_schedule_add_data_not_same_length(
    task_label, workers, start_times, time_lengths
):
    schedule = Schedule()

    with pytest.raises(ValueError):
        schedule.add_data(task_label, workers, start_times, time_lengths)


def test_schedule_fig_ax_attribute():
    workers, start_times, time_lengths = [1, 2], [1, 2], [3, 4]

    schedule = Schedule()
    schedule.add_data("data", workers, start_times, time_lengths)
    schedule.show()
    fig, ax = schedule.fig_ax

    assert type(fig) == figure.Figure
    assert type(ax) == axes.Subplot


def test_schedule_fig_ax_attribute_before_show():
    workers, start_times, time_lengths = [1, 2], [1, 2], [3, 4]

    schedule = Schedule()
    schedule.add_data("data", workers, start_times, time_lengths)

    with pytest.raises(AttributeError):
        schedule.fig_ax


def test_schedule_show_no_plot_data():
    schedule = Schedule()
    with pytest.raises(RuntimeError):
        schedule.show()


params = {
    "give argument": ("title", "title"),
    "default": (None, "schedule"),
}


@pytest.mark.parametrize(
    "title, expect",
    list(params.values()),
    ids=list(params.keys()),
)
def test_schedule_show_arg_title(title, expect):
    schedule = Schedule()
    schedule.add_data("data", [1, 2], [3, 4], [5, 6])
    schedule.show(title=title)
    fig, ax = schedule.fig_ax

    success_show_title = False
    for obj in fig.texts:
        actual_title = obj.get_text()
        if actual_title == expect:
            success_show_title = True

    assert success_show_title


def test_schedule_show_bar():
    workers1, start_times1, time_lengths1 = [1, 2], [1, 2], [3, 4]
    workers2, start_times2, time_lengths2 = [2], [1], [1]

    schedule = Schedule()
    schedule.add_data("data1", workers1, start_times1, time_lengths1)
    schedule.add_data("data2", workers2, start_times2, time_lengths2)

    schedule.show()
    fig, ax = schedule.fig_ax

    # Check that the children of ax contain the expected bar information
    expect_center_1 = np.array(
        [2.5, 1]
    )  # np.array([start_times1[0] + (time_length1[0] / 2), workers1[0]])
    success_show_bar_1 = False
    for obj in ax.get_children():
        if type(obj) == patches.Rectangle:
            actual_center_1 = obj.get_center()
            if (np.abs(actual_center_1 - expect_center_1) < 0.0001).all():
                success_show_bar_1 = True
    assert success_show_bar_1

    expect_center_2 = np.array(
        [4, 2]
    )  # np.array([start_times1[1] + (time_length1[1] / 2), workers1[1]])
    success_show_bar_2 = False
    for obj in ax.get_children():
        if type(obj) == patches.Rectangle:
            actual_center_2 = obj.get_center()
            if (np.abs(actual_center_2 - expect_center_2) < 0.0001).all():
                success_show_bar_2 = True
    assert success_show_bar_2

    expect_center_3 = np.array(
        [1.5, 2]
    )  # np.array([start_times2[0] + (time_length2[0] / 2), workers2[0]])
    success_show_bar_3 = False
    for obj in ax.get_children():
        if type(obj) == patches.Rectangle:
            actual_center_3 = obj.get_center()
            if (np.abs(actual_center_3 - expect_center_3) < 0.0001).all():
                success_show_bar_3 = True
    assert success_show_bar_3


def test_schedule_show_text():
    workers, start_times, time_lengths = [1, 2], [1, 2], [3, 4]

    schedule = Schedule()
    schedule.add_data("data", workers, start_times, time_lengths)
    schedule.show()
    fig, ax = schedule.fig_ax

    # Check that the ax.texts contain the expected text information
    expect_text_1 = "3"  # time_length[0]
    expect_center_1 = (2.5, 1)  # (start_times[0] + (time_length[0] / 2), workers[0])
    success_text_1 = False
    for obj in ax.texts:
        actual_text_1 = obj.get_text()
        actual_center_1 = obj.get_position()
        if not (actual_text_1 == expect_text_1):
            continue
        if not (actual_center_1 == expect_center_1):
            continue
        success_text_1 = True
    assert success_text_1

    expect_text_2 = "4"  # time_length[1]
    expect_center_2 = (4, 2)  # (start_times[1] + (time_length[1] / 2), workers[1])
    success_text_2 = False
    for obj in ax.texts:
        actual_text_2 = obj.get_text()
        actual_center_2 = obj.get_position()
        if not (actual_text_2 == expect_text_2):
            continue
        if not (actual_center_2 == expect_center_2):
            continue
        success_text_2 = True
    assert success_text_2


def test_schedule_show_arg_figsize():
    figwidth, figheight = 8, 4

    schedule = Schedule()
    schedule.add_data("data", [1, 2], [1, 2], [3, 4])
    schedule.show(figsize=tuple([figwidth, figheight]))
    fig, ax = schedule.fig_ax

    assert fig.get_figwidth() == 8
    assert fig.get_figheight() == 4


def test_schedule_show_arg_color_list():
    color_list = ["red", "blue"]
    workers1, start_times1, time_lengths1 = [1], [1], [3]
    workers2, start_times2, time_lengths2 = [2], [2], [4]

    schedule = Schedule()
    schedule.add_data("data1", workers1, start_times1, time_lengths1)
    schedule.add_data("data2", workers2, start_times2, time_lengths2)
    schedule.show(color_list=color_list)
    fig, ax = schedule.fig_ax

    # Check that the children of ax contain the expected color information
    expect_color_1 = np.array([1.0, 0.0, 0.0])  # red
    expect_center_1 = np.array(
        [2.5, 1]
    )  # np.array([start_times1[0] + (time_length1[0] / 2), workers1[0]])
    success_coloring_1 = False
    for obj in ax.get_children():
        if type(obj) == patches.Rectangle:
            actual_color_1 = obj.get_facecolor()[:-1]
            actual_center_1 = obj.get_center()
            if not (actual_color_1 == expect_color_1).all():
                continue
            if not (np.abs(actual_center_1 - expect_center_1) < 0.0001).all():
                continue
            success_coloring_1 = True
    assert success_coloring_1

    expect_color_2 = np.array([0.0, 0.0, 1.0])  # blue
    expect_center_2 = np.array(
        [4, 2]
    )  # np.array([start_times2[0] + (time_length2[0] / 2), workers2[0]])
    success_coloring_2 = False
    for obj in ax.get_children():
        if type(obj) == patches.Rectangle:
            actual_color_2 = obj.get_facecolor()[:-1]
            actual_center_2 = obj.get_center()
            if not (actual_color_2 == expect_color_2).all():
                continue
            if not (np.abs(actual_center_2 - expect_center_2) < 0.0001).all():
                continue
            success_coloring_2 = True
    assert success_coloring_2


def test_schedule_show_arg_color_list_invalid_length():
    color_list = ["r", "g", "b"]

    schedule = Schedule()
    schedule.add_data("data0", [1, 2], [1, 2], [3, 4])
    schedule.add_data("data1", [1, 2], [1, 2], [3, 4])

    with pytest.raises(ValueError):
        schedule.show(color_list=color_list)


params = {
    "give argument": ([0.3, 0.7], 0.3, 0.7),
    "default": (None, 0.5, 0.5),
}


@pytest.mark.parametrize(
    "alpha_list, expect_alpha_1, expect_alpha_2",
    list(params.values()),
    ids=list(params.keys()),
)
def test_schedule_show_arg_alpha_list(alpha_list, expect_alpha_1, expect_alpha_2):
    workers1, start_times1, time_lengths1 = [1], [1], [3]
    workers2, start_times2, time_lengths2 = [2], [2], [4]

    schedule = Schedule()
    schedule.add_data("data1", workers1, start_times1, time_lengths1)
    schedule.add_data("data2", workers2, start_times2, time_lengths2)
    schedule.show(alpha_list=alpha_list)
    fig, ax = schedule.fig_ax

    # Check that the children of ax contain the expected alpha information
    expect_center_1 = np.array(
        [2.5, 1]
    )  # np.array([start_times1[0] + (time_length1[0] / 2), workers1[0]])
    success_alpha_1 = False
    for obj in ax.get_children():
        if type(obj) == patches.Rectangle:
            actual_alpha_1 = obj.get_alpha()
            actual_center_1 = obj.get_center()
            if not actual_alpha_1 == expect_alpha_1:
                continue
            if not (np.abs(actual_center_1 - expect_center_1) < 0.0001).all():
                continue
            success_alpha_1 = True
    assert success_alpha_1

    expect_center_2 = np.array(
        [4, 2]
    )  # np.array([start_times1[0] + (time_length1[0] / 2), workers1[0]])
    success_alpha_2 = False
    for obj in ax.get_children():
        if type(obj) == patches.Rectangle:
            actual_alpha_2 = obj.get_alpha()
            actual_center_2 = obj.get_center()
            if not actual_alpha_2 == expect_alpha_2:
                continue
            if not (np.abs(actual_center_2 - expect_center_2) < 0.0001).all():
                continue
            success_alpha_2 = True
    assert success_alpha_2


def test_schedule_show_arg_alpha_list_invalid_length():
    alpha_list = [0.1, 0.1, 0.1]

    schedule = Schedule()
    schedule.add_data("data0", [1, 2], [1, 2], [3, 4])
    schedule.add_data("data1", [1, 2], [1, 2], [3, 4])

    with pytest.raises(ValueError):
        schedule.show(alpha_list=alpha_list)


params = {
    "give argument": ("xlabel", "xlabel"),
    "default": (None, "time"),
}


@pytest.mark.parametrize(
    "xlabel, expect",
    list(params.values()),
    ids=list(params.keys()),
)
def test_schedule_show_arg_xlabel(xlabel, expect):
    schedule = Schedule()
    schedule.add_data("data", [1, 2], [1, 2], [3, 4])
    schedule.show(xlabel=xlabel)
    fig, ax = schedule.fig_ax

    assert ax.get_xlabel() == expect


params = {
    "give argument": ("ylabel", "ylabel"),
    "default": (None, "worker"),
}


@pytest.mark.parametrize(
    "ylabel, expect",
    list(params.values()),
    ids=list(params.keys()),
)
def test_schedule_show_arg_ylabel(ylabel, expect):
    schedule = Schedule()
    schedule.add_data("data", [1, 2], [1, 2], [3, 4])
    schedule.show(ylabel=ylabel)
    fig, ax = schedule.fig_ax

    assert ax.get_ylabel() == expect


def test_schedule_show_arg_xticks():
    xticks = [1, 2, 3, 4, 5, 6]

    schedule = Schedule()
    schedule.add_data("data", [1, 2], [1, 2], [3, 4])
    schedule.show(xticks=xticks)
    fig, ax = schedule.fig_ax

    assert (ax.get_xticks() == np.array([1, 2, 3, 4, 5, 6])).all()


params = {
    "give argument": ([1, 2, 3], np.array([1, 2, 3])),
    "default": (None, np.array([1, 2])),
}


@pytest.mark.parametrize(
    "yticks, expect",
    list(params.values()),
    ids=list(params.keys()),
)
def test_schedule_show_arg_yticks(yticks, expect):
    schedule = Schedule()
    schedule.add_data("data", [1, 2], [1, 2], [3, 4])
    schedule.show(yticks=yticks)
    fig, ax = schedule.fig_ax

    assert (ax.get_yticks() == expect).all()


params = {
    "undirected case": ([[0, 1], [1, 2]], GraphType.UNDIRECTED, nx.Graph),
    "directed case": ([[0, 1], [1, 2]], GraphType.DIRECTED, nx.DiGraph),
}


@pytest.mark.parametrize(
    "edge_list, graphtype, expect_type",
    list(params.values()),
    ids=list(params.keys()),
)
def test_graph_from_edge_list(edge_list, graphtype, expect_type):
    graph = Graph.from_edge_list(edge_list, graphtype)
    G = graph.G

    assert type(G) == expect_type
    assert len(G.edges()) == 2


params = {
    "undirected case": ([[-1, 1], [1, -1]], GraphType.UNDIRECTED, nx.Graph, 1),
    "directed case": ([[-1, 1], [2, -1]], GraphType.DIRECTED, nx.DiGraph, 2),
    "numpy case": (np.array([[-1, 1], [1, -1]]), GraphType.UNDIRECTED, nx.Graph, 1),
}


@pytest.mark.parametrize(
    "distance_matrix, graphtype, expect_type, expect_edge_num",
    list(params.values()),
    ids=list(params.keys()),
)
def test_graph_from_distance_matrix(
    distance_matrix, graphtype, expect_type, expect_edge_num
):
    graph = Graph.from_distance_matrix(distance_matrix, graphtype)
    G = graph.G

    assert type(G) == expect_type
    assert len(G.edges()) == expect_edge_num


def test_graph_fig_ax_attribute():
    graph = Graph.from_edge_list([[1, 2]], GraphType.UNDIRECTED)
    graph.show()
    fig, ax = graph.fig_ax

    assert type(fig) == figure.Figure
    assert type(ax) == axes.Subplot


def test_graph_fig_ax_attribute_before_show():
    graph = Graph.from_edge_list([[1, 2]], GraphType.UNDIRECTED)

    with pytest.raises(AttributeError):
        graph.fig_ax


params = {
    "give argument": ("title", "title"),
    "default": (None, "graph"),
}


@pytest.mark.parametrize(
    "title, expect",
    list(params.values()),
    ids=list(params.keys()),
)
def test_graph_show_arg_title(title, expect):
    graph = Graph.from_edge_list([[1, 2]], GraphType.UNDIRECTED)
    graph.show(title=title)
    fig, ax = graph.fig_ax

    # Check that the ax.texts contain the expected node label information
    success_show_title = False
    for obj in fig.texts:
        actual_title = obj.get_text()
        if actual_title == expect:
            success_show_title = True

    assert success_show_title


def test_graph_show_arg_figsize():
    figwidth, figheight = 8, 4

    graph = Graph.from_edge_list([[1, 2]], GraphType.UNDIRECTED)
    graph.show(figsize=tuple([figwidth, figheight]))
    fig, ax = graph.fig_ax

    assert fig.get_figwidth() == 8
    assert fig.get_figheight() == 4


def test_graph_show_node():
    graph = Graph.from_edge_list([[1, 2], [2, 3]], GraphType.UNDIRECTED)
    graph.show()
    fig, ax = graph.fig_ax

    # Check that the children of ax contain the expected node information
    expect_node_num = 3
    success_show_node = False
    for obj in ax.get_children():
        if type(obj) == collections.PathCollection:
            actual_node_num = obj.get_offsets().data.shape[0]
            if (np.abs(actual_node_num - expect_node_num) < 0.0001).all():
                success_show_node = True
    assert success_show_node


def test_graph_show_arg_node_pos():
    pos1 = np.array([1, 1])
    pos2 = np.array([-1, -1])
    node_pos = {1: pos1, 2: pos2}

    # Check that the children of ax contain the expected node position information
    graph = Graph.from_edge_list([[1, 2]], GraphType.UNDIRECTED)
    graph.show(node_pos=node_pos)
    fig, ax = graph.fig_ax

    expect_pos = np.vstack([pos1, pos2])
    success_set_pos = False
    for obj in ax.get_children():
        if type(obj) == collections.PathCollection:
            actual_pos = obj.get_offsets().data
            if (np.abs(actual_pos - expect_pos) < 0.0001).all():
                success_set_pos = True
    assert success_set_pos


def test_graph_show_arg_node_pos_default():
    graph = Graph.from_edge_list([[1, 2]], GraphType.UNDIRECTED)
    graph.show()
    fig, ax = graph.fig_ax

    # Check that the children of ax contain the expected node position information
    default_pos = nx.spring_layout(graph.G, seed=1)
    expect_pos = np.vstack(list(default_pos.values()))
    success_set_pos = False
    for obj in ax.get_children():
        if type(obj) == collections.PathCollection:
            actual_pos = obj.get_offsets().data
            if (np.abs(actual_pos - expect_pos) < 0.0001).all():
                success_set_pos = True
    assert success_set_pos


def test_graph_show_arg_node_color():
    node_color = ["r", "b"]

    graph = Graph.from_edge_list([[1, 2]], GraphType.UNDIRECTED)
    graph.show(node_color=node_color)
    fig, ax = graph.fig_ax

    # Check that the children of ax contain the expected node color information
    expect_color_node1 = np.array(matplotlib.colors.to_rgb("r"))
    success_coloring_node1 = False
    for obj in ax.get_children():
        if type(obj) == collections.PathCollection:
            actual_color_node1 = obj.get_facecolor()[0][:-1]
            if (np.abs(actual_color_node1 - expect_color_node1) < 0.0001).all():
                success_coloring_node1 = True
    assert success_coloring_node1

    expect_color_node2 = np.array(matplotlib.colors.to_rgb("b"))
    success_coloring_node2 = False
    for obj in ax.get_children():
        if type(obj) == collections.PathCollection:
            actual_color_node2 = obj.get_facecolor()[1][:-1]
            if (np.abs(actual_color_node2 - expect_color_node2) < 0.0001).all():
                success_coloring_node2 = True
    assert success_coloring_node2


def test_graph_show_arg_node_color_default():
    default_color = "#1f78b4"

    graph = Graph.from_edge_list([[1, 2]], GraphType.UNDIRECTED)
    graph.show()
    fig, ax = graph.fig_ax

    expect_color = np.array(matplotlib.colors.to_rgb(default_color))

    # Check that the children of ax contain the expected node color information
    success_coloring = False
    for obj in ax.get_children():
        if type(obj) == collections.PathCollection:
            actual_color = obj.get_facecolor()[0][:-1]
            if (np.abs(actual_color - expect_color) < 0.0001).all():
                success_coloring = True
    assert success_coloring


def test_graph_show_arg_node_labels():
    node1, node2 = 1, 2
    node_labels = {node1: "node1", node2: "node2"}
    node_pos = {node1: np.array([1, 1]), node2: np.array([-1, -1])}

    graph = Graph.from_edge_list([[1, 2]], GraphType.UNDIRECTED)
    graph.show(node_pos=node_pos, node_labels=node_labels)
    fig, ax = graph.fig_ax

    # Check that the ax.texts contain the expected node label information
    expect_info_node1 = (node_pos[node1][0], node_pos[node1][1], "node1")
    success_show_node1_label = False
    for obj in ax.texts:
        actual_info_node1 = (*obj.get_position(), obj.get_text())
        if actual_info_node1 == expect_info_node1:
            success_show_node1_label = True
    assert success_show_node1_label

    expect_info_node2 = (node_pos[node2][0], node_pos[node2][1], "node2")
    success_show_node2_label = False
    for obj in ax.texts:
        actual_info_node2 = (*obj.get_position(), obj.get_text())
        if actual_info_node2 == expect_info_node2:
            success_show_node2_label = True
    assert success_show_node2_label


def test_graph_show_arg_node_labels_default():
    node1, node2 = 1, 2
    node_pos = {node1: np.array([1, 1]), node2: np.array([-1, -1])}

    graph = Graph.from_edge_list([[node1, node2]], GraphType.UNDIRECTED)
    graph.show(node_pos=node_pos)
    fig, ax = graph.fig_ax

    # Check that the ax.texts contain the expected node label information
    expect_info_node1 = (node_pos[node1][0], node_pos[node1][1], str(node1))
    success_show_node1_label = False
    for obj in ax.texts:
        actual_info_node1 = (*obj.get_position(), obj.get_text())
        if actual_info_node1 == expect_info_node1:
            success_show_node1_label = True
    assert success_show_node1_label

    expect_info_node2 = (node_pos[node2][0], node_pos[node2][1], str(node2))
    success_show_node2_label = False
    for obj in ax.texts:
        actual_info_node2 = (*obj.get_position(), obj.get_text())
        if actual_info_node2 == expect_info_node2:
            success_show_node2_label = True
    assert success_show_node2_label


def test_graph_show_arg_edge():
    edge_list = [[1, 2], [2, 3]]

    graph = Graph.from_edge_list(edge_list, GraphType.UNDIRECTED)
    graph.show()
    fig, ax = graph.fig_ax

    expect_edge_num = 2

    # Check that the children of ax contain the expected edge information
    success_show_edge = False
    for obj in ax.get_children():
        if type(obj) == collections.LineCollection:
            actual_edge_num = len(obj.get_paths())
            if actual_edge_num == expect_edge_num:
                success_show_edge = True
    assert success_show_edge


def test_graph_show_arg_edge_color():
    edge_color = ["r"]

    graph = Graph.from_edge_list([[1, 2]], GraphType.UNDIRECTED)
    graph.show(edge_color=edge_color)
    fig, ax = graph.fig_ax

    expect_color = np.array([1.0, 0.0, 0.0])

    # Check that the children of ax contain the expected edge color information
    success_coloring = False
    for obj in ax.get_children():
        if type(obj) == collections.LineCollection:
            actual_color = obj.get_color()[0][:-1]
            if (actual_color == expect_color).all():
                success_coloring = True
    assert success_coloring


def test_graph_show_arg_edge_color_default():
    graph = Graph.from_edge_list([[1, 2]], GraphType.UNDIRECTED)
    graph.show()
    fig, ax = graph.fig_ax

    expect_color = np.array([0.0, 0.0, 0.0])  # "black"

    # Check that the children of ax contain the expected edge color information
    success_coloring = False
    for obj in ax.get_children():
        if type(obj) == collections.LineCollection:
            actual_color = obj.get_color()[0][:-1]
            if (actual_color == expect_color).all():
                success_coloring = True
    assert success_coloring


def test_graph_show_arg_edge_labels():
    node_pos = {
        0: np.array([-1, -1]),
        1: np.array([1, 1]),
    }
    edge_labels = {(0, 1): "edge"}

    graph = Graph.from_edge_list([[0, 1]], GraphType.UNDIRECTED)
    graph.show(node_pos=node_pos, edge_labels=edge_labels)
    fig, ax = graph.fig_ax

    # Check that the ax.texts contain the expected edge label information
    expect_label_x, expect_label_y = (node_pos[0] + node_pos[1]) / 2
    expect_label = edge_labels[(0, 1)]
    expect_info = (expect_label_x, expect_label_y, expect_label)

    success_show_edge_label = False
    for obj in ax.texts:
        actual_info = (*obj.get_position(), obj.get_text())
        if actual_info == expect_info:
            success_show_edge_label = True

    assert success_show_edge_label


def test_graph_show_arg_edge_labels_default_weighted_edge_case():
    node_pos = {0: np.array([1, 1]), 1: np.array([-1, -1])}
    weight = 5

    graph = Graph.from_distance_matrix(
        [[-1, weight], [weight, -1]], GraphType.UNDIRECTED
    )
    graph.show(node_pos=node_pos)
    fig, ax = graph.fig_ax

    # Check that the ax.texts contain the expected edge label information
    expect_label_x, expect_label_y = (node_pos[0] + node_pos[1]) / 2
    expect_label = str(weight)
    expect_info = (expect_label_x, expect_label_y, expect_label)

    success_show_edge_label = False
    for obj in ax.texts:
        actual_info = (*obj.get_position(), obj.get_text())
        if actual_info == expect_info:
            success_show_edge_label = True

    assert success_show_edge_label



from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)


from jijbench.datasets.problem import (
    TSP,
    TSPTW,
    BinPacking,
    Knapsack,
    NurseScheduling,
    # get_problem,
)

# from jijbench.datasets.instance_data import get_instance_data
#
__all__ = [
    "BinPacking",
    "Knapsack",
    "TSP",
    "TSPTW",
    "NurseScheduling",
]



from __future__ import annotations

import typing as tp
import inspect
import jijmodeling as jm

from dataclasses import dataclass
from jijbench.exceptions.exceptions import SolverFailedError
from jijbench.node.base import DataNode, FunctionNode
from jijbench.containers.containers import Record
from jijbench.functions.factory import RecordFactory
from jijbench.typing import T


@dataclass
class Parameter(DataNode[T]):
    """A parameter for a solver function.

    Attributes:
        data (Any): The data in the node.
        name (str): The name of the parameter.
    """

    name: str

    @classmethod
    def validate_data(cls, data: tp.Any) -> tp.Any:
        """A class method to validate the data before setting it.

        Args:
            data (Any): The data to be validated.

        Returns:
            Any: The validated data.
        """
        return data


@dataclass
class Response(DataNode[T]):
    """A return value of a solver function.

    Attributes:
        data (Any): The data in the node.
        name (str): The name of the return value.
    """

    name: str

    @classmethod
    def validate_data(cls, data: tp.Any) -> tp.Any:
        """A class method to validate the data before setting it.

        Args:
            data (Any): The data to be validated.

        Returns:
            Any: The validated data.
        """
        return data


class Solver(FunctionNode[Parameter, Record]):
    """A solver function that takes a list of Parameter and returns a Record.

    Attributes:
        name (str): The name of the solver function.
        function (Callable): The actual function to be executed.
    """

    def __init__(self, function: tp.Callable, name: str | None = None) -> None:
        """The constructor of the `Solver` class.

        Args:
            function (Callable): The actual function to be executed.
            name (str, optional): The name of the solver function. Defaults to None.
        """
        if name is None:
            name = function.__name__
        super().__init__(name)
        self.function = function

    def operate(
        self,
        inputs: list[Parameter],
    ) -> Record:
        """The main operation of the solver function.

        Args:
            inputs (list[Parameter]): The list of input `Parameter` for the solver function.
            is_parsed_sampleset (bool, optional): Whether the sample set is parsed. Defaults to True.

        Raises:
            SolverFailedError: If an error occurs inside the solver function.

        Returns:
            Record: The result of the solver function as a `Record`.
        """
        from jijbench.solver.jijzept import SampleSet

        parameters = inspect.signature(self.function).parameters
        solver_args = {
            node.name: node.data for node in inputs if node.name in parameters
        }
        try:
            rets = self.function(**solver_args)
            if not isinstance(rets, tuple):
                rets = (rets,)
        except Exception as e:
            msg = f'An error occurred inside your solver. Please check implementation of "{self.name}". -> {e}'
            raise SolverFailedError(msg)

        solver_return_names = [f"{self.name}_return[{i}]" for i in range(len(rets))]

        rets = [
            SampleSet(data, name)
            if isinstance(data, jm.SampleSet)
            else Response(data, name)
            for data, name in zip(rets, solver_return_names)
        ]
        factory = RecordFactory()
        return factory(rets)



import pytest

import jijbench as jb
import jijmodeling as jm
import numpy as np
import pytest

from jijbench.exceptions.exceptions import SolverFailedError


def func1(x):
    return x


OBJECT = "OBJECT"
UNSUPPORTED_SOLVER = 0


def custom_solver_failed():
    raise Exception("solver is failed.")


def test_simple_solver():
    solver = jb.Solver(func1)

    param = jb.Parameter(1, "x")
    record = solver([param])

    assert isinstance(record, jb.Record)
    assert record.data[0].data == 1
    assert record.operator is None


def test_CallebleSolver_solver_failed_error():
    solver = jb.Solver(custom_solver_failed)
    with pytest.raises(SolverFailedError):
        solver([])


def test_instance_data():
    instance_data = jb.InstanceData(
        {
            "int": 1,
            "float": 1.0,
            "list": [1, 2.0],
            "ndarray": np.array([[1, 2], [3, 4]]),
        },
        "sample",
    )

    assert instance_data.data["int"] == 1
    assert instance_data.data["float"] == 1.0
    assert instance_data.data["list"] == [1, 2.0]
    assert isinstance(instance_data.data["ndarray"], np.ndarray)


def test_invalid_instance_data_keys():
    with pytest.raises(TypeError):
        instance_data = jb.InstanceData(
            {
                "int": 1,
                "float": 1.0,
                0: [1, 2.0],
                1: np.array([[1, 2], [3, 4]]),
            },
            "sample",
        )


def test_invalid_instance_data_values():
    with pytest.raises(TypeError):
        instance_data = jb.InstanceData(
            {
                "int": "a",
                "float": {"b": 1},
                "list": [1, 2.0],
                "ndarray": np.array([[1, 2], [3, 4]]),
            },
            "sample",
        )


def test_user_defined_model():
    problem = jm.Problem("sample")
    a = jm.Placeholder("a")
    b = jm.Placeholder("b")
    c = jm.Placeholder("c", 1)
    i = jm.Element("i", 5)
    x = jm.DecisionVariable("x", 5)

    problem = jm.Problem("sample")
    problem += a + jm.Sum(i, c[i] * x[i])
    problem += jm.Constraint("const", x[:] == b)

    instance_data: jm.PH_VALUES_INTERFACE = {"a": 1, "b": 2.0, "c": [1, 2]}

    model = jb.UserDefinedModel((problem, instance_data), "test")


def test_invalid_user_defined_model():
    problem = jm.Problem("sample")
    a = jm.Placeholder("a")
    b = jm.Placeholder("b")
    c = jm.Placeholder("c", 1)
    i = jm.Element("i", 5)
    x = jm.DecisionVariable("x", 5)

    problem = jm.Problem("sample")
    problem += a + jm.Sum(i, c[i] * x[i])
    problem += jm.Constraint("const", x[:] == b)

    instance_data: jm.PH_VALUES_INTERFACE = {"a": 1, "c": [1, 2]}
    with pytest.raises(KeyError):
        model = jb.UserDefinedModel((problem, instance_data), "test")



from __future__ import annotations

import jijmodeling as jm
from numbers import Number
import numpy as np
import pandas as pd
import typing as tp

from jijbench.experiment.experiment import Experiment
from jijbench.functions.concat import Concat
from jijbench.solver.base import Parameter, Response
from jijbench.functions.factory import RecordFactory


def construct_experiment_from_samplesets(
    samplesets: list[jm.SampleSet] | jm.SampleSet,
    additional_data: dict[str, list[tp.Any]] | None = None,
) -> Experiment:
    """Construct `jb.Experiment` instance from a list of `jm.SampleSet`.

    The visualization function of JijBenchmark is implemented for `jb.Experiment`.
    These function can be applied to the user's `jm.SampleSet` through this function.

    Args:
        samplesets (list[jm.SampleSet] | jm.SampleSet): a list of JijModeling SampleSet. You can also just give a single `jm.SampleSet`.
        additional_data (dict[str, list[tp.Any]] | None):  a dictionary of data to store in the experiment.
            The key will be the jb.Experiment.table column name and the value is the list of elements stored in the table.
            The length of this list must equal the length of samplesets list.
            Defaults to None.

    Returns:
        Experiment: a `jb.Experiment` instance. The number of rows in `jb.Experiment.table` is equal to the length of samplesets.

    Example:
        The code below solves the TSP problem and gets the jb.Experiment from that sampleset.

        ```python
        import jijbench as jb
        import jijzept as jz
        from jijbench.visualization.metrics.utils import construct_experiment_from_samplesets

        problem = jb.get_problem("TSP")
        instance_data = jb.get_instance_data("TSP")[0][1]

        # config_path = "XX"
        sampler = jz.JijSASampler(config=config_path)

        samplesets = []
        onehot_time_multipliers = []
        onehot_location_multipliers = []

        for onehot_time_multiplier in [0.01, 0.1]:
            for onehot_location_multiplier in [0.01, 0.1]:
                multipliers = {"onehot_time": onehot_time_multiplier, "onehot_location": onehot_location_multiplier}
                sampleset = sampler.sample_model(
                    model=problem,
                    feed_dict=instance_data,
                    multipliers=multipliers,
                    num_reads=10,
                )
                samplesets.append(sampleset)
                onehot_time_multipliers.append(onehot_time_multiplier)
                onehot_location_multipliers.append(onehot_location_multiplier)

        additional_data = {
            "onehot_time_multiplier": onehot_time_multipliers,
            "onehot_location_multiplier": onehot_location_multipliers,
        }
        experiment = construct_experiment_from_samplesets(samplesets, additional_data)
        ```
    """
    if isinstance(samplesets, jm.SampleSet):
        samplesets = [samplesets]

    if additional_data is None:
        additional_data = {}
    else:
        for v in additional_data.values():
            if len(v) != len(samplesets):
                raise TypeError(
                    "The list assigned to the value of additional_data must have the same length as the sampleset."
                )

    # Convert additional_data to JijBenchmark Parameters.
    params = [
        [
            v if isinstance(v, Parameter) else Parameter(v, k)
            for k, v in zip(additional_data.keys(), r)
        ]
        for r in zip(*additional_data.values())
    ]
    experiment = Experiment(autosave=False)
    for i, sampleset in enumerate(samplesets):
        factory = RecordFactory()
        ret = [Response(data=sampleset, name="")]
        record = factory(ret)
        # Concat additional_data if given.
        if len(params) >= 1:
            record = Concat()([RecordFactory()(params[i]), record])
        experiment.append(record)
    return experiment


def _create_fig_title_list(
    metrics: pd.Series,
    title: str | list[str] | None,
) -> list[str]:
    """Create figure title list for Visualization, each title is passed to `matplotlib.pyplot.suptitle`.

    This function produces a title list of length equal to the number of rows in the metrics series.
    JijBenchmark`s metrics plot draws a figure for each run (i.e. each row of `jb.Experiment.table`),
    and each element of the returned list is expected to be the suptitle of each figure.

    Args:
        metrics (pd.Series): A `pd.Series` instance of the metrics for each run.
        title (str | list[str] | None): A title, or a `list` of titles. If `None`, the title list is automatically generated from the metrics indices.

    Returns:
        list[str]: a list of the suptitle of the figure. Its length is equal to the number of rows in the metrics series.
    """
    if isinstance(title, list):
        title_list = title
        return title_list
    elif isinstance(title, str):
        title_list = [title for _ in range(len(metrics))]
        return title_list
    elif title is None:
        title_list = []
        index_names = metrics.index.names
        for index, _ in metrics.items():
            # If user don't give title, the title list is automatically generated from the metrics index.
            if index is None:
                title_list.append("")
            elif isinstance(index, tuple):
                title_list.append(
                    "\n".join(
                        [
                            f"{index_name}: {index_element}"
                            for index_name, index_element in zip(index_names, index)
                        ]
                    )
                )
            else:
                index_name = index_names[0] if index_names[0] is not None else "index"
                title_list.append(f"{index_name}: {index}")
        return title_list
    else:
        raise TypeError("title must be str or list[str].")


def _df_has_valid_multipliers_column(df: pd.DataFrame) -> bool:
    """
    Check that the `pd.DataFrame` instance has `multipliers` column in `JijBenchmark` format.
    """
    if "multipliers" not in df.columns:
        return False

    def element_is_valid(x: pd.Series, constraint_names: list[str]) -> bool:
        multipliers = x["multipliers"]
        if not isinstance(multipliers, dict):
            return False
        for key, value in multipliers.items():
            if not isinstance(key, str):
                return False
            if not isinstance(value, Number):
                return False
        if list(multipliers.keys()) != constraint_names:
            return False
        return True

    first_element = df["multipliers"].values[0]
    if isinstance(first_element, dict):
        constraint_names = list(first_element.keys())
    else:
        return False
    check_results = df.apply(
        element_is_valid,
        axis=1,
        constraint_names=constraint_names,
    )
    return all(check_results.values)


def _df_has_number_array_column_target_name(df: pd.DataFrame, column_name: str) -> bool:
    """
    Check that the `pd.DataFrame` instance has a column named `column_name` and its element is number array.
    """
    if column_name not in df.columns:
        return False

    def element_is_number_array(x: pd.Series, column_name: str) -> bool:
        element = x[column_name]
        if not isinstance(element, (list, np.ndarray)):
            return False
        if np.array_equal(element, np.asarray(None)):
            return False
        for num in element:
            if not isinstance(num, Number):
                return False
        return True

    check_results = df.apply(
        element_is_number_array,
        axis=1,
        column_name=column_name,
    )
    return all(check_results.values)


def _df_has_number_column_target_name(df: pd.DataFrame, column_name: str) -> bool:
    """
    Check that the `pd.DataFrame` instance has a column named `column_name` and its element is number.
    """
    if column_name not in df.columns:
        return False

    def element_is_number(x: pd.Series, column_name: str) -> bool:
        element = x[column_name]
        if not isinstance(element, Number):
            return False
        return True

    check_results = df.apply(
        element_is_number,
        axis=1,
        column_name=column_name,
    )
    return all(check_results.values)



from __future__ import annotations

import jijmodeling as jm
import numpy as np

from dataclasses import dataclass
from jijbench.solver.base import Parameter, Response
from jijbench.typing import ModelType
from jijmodeling.expression.extract import extract_vars_from_problem


@dataclass
class InstanceData(Parameter[jm.PH_VALUES_INTERFACE]):
    """Instance data for JijZept.

    Attributes:
        data (jijmodeling.PH_VALUES_INTERFACE): A dictionary where keys are the string labels of placeholders in the optimization problem,
            and values are integer, float, list, or numpy.ndarray type data.
        name (str): The name of instance_data.
    """

    @classmethod
    def validate_data(cls, data: jm.PH_VALUES_INTERFACE) -> jm.PH_VALUES_INTERFACE:
        """
        Validate the instance data to ensure it has the correct format.

        Args:
            data (dict): The instance data to validate.

        Raises:
            TypeError: If any key in the data is not of type str.
            TypeError: If any value in the data is not of type int, float, list, or numpy.ndarray.

        Returns:
            dict: The validated instance data.
        """
        data = cls._validate_dtype(data, (dict,))

        is_instance_data_keys = [isinstance(k, str) for k in data]
        if not all(is_instance_data_keys):
            invalid_keys = [k for b, k in zip(is_instance_data_keys, data) if not b]
            raise TypeError(
                f"The following key(s) {invalid_keys} of instance data is invalid. The type of key must be str."
            )

        is_instance_data_values = [
            isinstance(v, (int, float, list, np.ndarray)) for v in data.values()
        ]
        if not all(is_instance_data_values):
            invalid_values = [
                v for b, v in zip(is_instance_data_values, data.values()) if not b
            ]
            raise TypeError(
                f"The following value(s) {invalid_values} of instance data is invalid. The type of value must be int, float, list, or numpy.ndarray."
            )
        return data


@dataclass
class UserDefinedModel(Parameter[ModelType]):
    """User defined model for jijzept.

    Attributes:
        data: A tuple of `Problem` and `PH_VALUES_INTERFACE` which contains the
            problem definition and the instance data.
        name (str): The name of model.
    """

    @classmethod
    def validate_data(cls, data: ModelType) -> ModelType:
        """Validate the instance data in user-defined model.

        Args:
            data: A tuple of `Problem` and `PH_VALUES_INTERFACE` which contains the problem and the instance data.

        Raises:
            KeyError: If any label in the `Problem` is missing from the instance data.

        Returns:
            ModelType: The validated data.
        """
        problem, instance_data = data
        keys = list(instance_data.keys())
        ph_labels = [
            v.label
            for v in extract_vars_from_problem(problem)
            if isinstance(v, jm.Placeholder)
        ]

        is_in_labels = list(map(lambda x: x in keys, ph_labels))
        if not all(is_in_labels):
            missing_labels = [p for b, p in zip(is_in_labels, ph_labels) if not b]
            raise KeyError(
                f"Instance data needs label(s) {missing_labels}, but are not included."
            )
        return data

    @property
    def problem(self) -> jm.Problem:
        """Return the problem in the data of user defined model."""
        return self.data[0]

    @property
    def instance_data(self) -> jm.PH_VALUES_INTERFACE:
        """Return the instance data in the data of user defined model."""
        return self.data[1]


@dataclass
class SampleSet(Response[jm.SampleSet]):
    @classmethod
    def validate_data(cls, data: jm.SampleSet) -> jm.SampleSet:
        return cls._validate_dtype(data, (jm.SampleSet,))

    @property
    def record(self) -> jm.Record:
        return self.data.record

    @property
    def evaluation(self) -> jm.Evaluation:
        return self.data.evaluation

    @property
    def time(self) -> jm.MeasuringTime:
        return self.data.measuring_time



import dill
import jijbench as jb
import pandas as pd
import pathlib
import pytest
import shutil


SAVEDIR = "./jb_results"


@pytest.fixture(scope="function", autouse=True)
def pre_post_process():
    # preprocess
    p = pathlib.Path(SAVEDIR)
    p.mkdir(exist_ok=True)
    yield
    # postprocess
    p = pathlib.Path(SAVEDIR)
    if p.exists():
       shutil.rmtree(p)


def test_save_artifact_with_mode_w():
    a = jb.Artifact({0: {"a": jb.Number(1, "num")}})
    jb.save(a, savedir=SAVEDIR)

    with open(f"{SAVEDIR}/artifact.dill", "rb") as f:
        res = dill.load(f)

    assert res == a


def test_save_artifact_with_mode_a():
    a0 = jb.Artifact({0: {"a": jb.Number(1, "num")}})
    jb.save(a0, savedir=SAVEDIR)

    a1 = jb.Artifact({1: {"a": jb.Number(1, "num")}})
    jb.save(a1, savedir=SAVEDIR, mode="a")

    with open(f"{SAVEDIR}/artifact.dill", "rb") as f:
        res = dill.load(f)

    assert res.data[0] == a0.data[0]
    assert res.data[1] == a1.data[1]


def test_load_artifact():
    a0 = jb.Artifact({0: {"a": jb.Number(1, "num")}})
    jb.save(a0, savedir=SAVEDIR)

    a1 = jb.Artifact({1: {"a": jb.Number(1, "num")}})
    jb.save(a1, savedir=SAVEDIR, mode="a")

    res = jb.load(SAVEDIR, return_type="Artifact")
    assert res.data[0] == a0.data[0]
    assert res.data[1] == a1.data[1]


def test_save_table_with_mode_w():
    t = jb.Table(pd.DataFrame([[jb.Number(1, "num")]], index=["a"]))
    jb.save(t, savedir=SAVEDIR)

    res = pd.read_csv(f"{SAVEDIR}/table.csv", index_col=0)
    with open(f"{SAVEDIR}/meta.dill", "rb") as f:
        meta = dill.load(f)

    assert res.loc["a", "0"] == 1
    assert meta["dtype"][0] == jb.Number


def test_save_table_with_mode_a():
    t0 = jb.Table(pd.DataFrame([[jb.Number(1, "num")]], index=["a"]))
    jb.save(t0, savedir=SAVEDIR)

    t1 = jb.Table(pd.DataFrame([[jb.Number(1, "num")]], index=["b"]))
    jb.save(t1, savedir=SAVEDIR, mode="a")

    res = pd.read_csv(f"{SAVEDIR}/table.csv", index_col=0)
    with open(f"{SAVEDIR}/meta.dill", "rb") as f:
        meta = dill.load(f)

    assert res.loc["a", "0"] == 1
    assert res.loc["b", "0"] == 1
    assert meta["dtype"][0] == jb.Number


def test_load_table():
    t0 = jb.Table(pd.DataFrame([[jb.Number(1, "num")]], index=["a"]))
    jb.save(t0, savedir=SAVEDIR)

    t1 = jb.Table(pd.DataFrame([[jb.Number(1, "num")]], index=["b"]))
    jb.save(t1, savedir=SAVEDIR, mode="a")

    res = jb.load(SAVEDIR, return_type="Table")
    assert res.data.loc["a", 0] == t0.data.loc["a", 0]
    assert res.data.loc["b", 0] == t1.data.loc["b", 0]


def test_save_experiment_with_mode_w():
    experiment_name = "0"
    a = jb.Artifact({0: {"a": jb.Number(1, "num")}})
    t = jb.Table(pd.DataFrame([[jb.Number(1, "num")]], index=["a"]))
    e = jb.Experiment((a, t), name=experiment_name)
    jb.save(e, savedir=SAVEDIR, mode="w")


def test_save_experiment_with_mode_a():
    experiment_name = "0"
    a0 = jb.Artifact({0: {"a": jb.Number(1, "num")}})
    t0 = jb.Table(pd.DataFrame([[jb.Number(1, "num")]], index=["a"]))
    e0 = jb.Experiment((a0, t0), name=experiment_name)
    jb.save(e0, savedir=SAVEDIR, mode="w")

    a1 = jb.Artifact({1: {"a": jb.Number(1, "num")}})
    t1 = jb.Table(pd.DataFrame([[jb.Number(1, "num")]], index=["b"]))
    e1 = jb.Experiment((a1, t1), name=experiment_name)
    jb.save(e1, savedir=SAVEDIR, mode="a")


def test_load_experiment():
    experiment_name = "0"
    a0 = jb.Artifact({0: {"a": jb.Number(1, "num")}})
    t0 = jb.Table(pd.DataFrame([[jb.Number(1, "num")]], index=["a"]))
    e0 = jb.Experiment((a0, t0), name=experiment_name)
    jb.save(e0, savedir=SAVEDIR, mode="w")

    a1 = jb.Artifact({1: {"a": jb.Number(1, "num")}})
    t1 = jb.Table(pd.DataFrame([[jb.Number(1, "num")]], index=["b"]))
    e1 = jb.Experiment((a1, t1), name=experiment_name)
    jb.save(e1, savedir=SAVEDIR, mode="a")

    res = jb.load(SAVEDIR, experiment_names=[experiment_name])

    artifact, table = res.data
    assert artifact.data[0] == a0.data[0]
    assert artifact.data[1] == a1.data[1]
    assert table.data.loc["a", 0] == t0.data.loc["a", 0]
    assert table.data.loc["b", 0] == t1.data.loc["b", 0]


def test_load_benchmark_results():
    def func(x):
        return x

    benchmark_name = "test"
    bench = jb.Benchmark({"x": [1, 2]}, solver=func, name=benchmark_name)
    res = bench(savedir=SAVEDIR)

    res_l = jb.load(benchmark_name, savedir=SAVEDIR)



from __future__ import annotations

import numpy as np


from dataclasses import dataclass
from jijbench.elements.base import Number
from jijbench.functions.math import Min, Max, Mean, Std
from jijbench.node.base import DataNode


@dataclass
class Array(DataNode[np.ndarray]):
    """A class representing numpy arrays.
    
    Attributes:
        data (numpy.ndarray): The numpy array.
        name (str): The name of the node.
    """
    def min(self) -> Number:
        """Get the minimum value of the numpy array.

        Returns:
            Number: The minimum value of the numpy array.
        """
        return self.apply(Min())

    def max(self) -> Number:
        """Get the maximum value of the numpy array.

        Returns:
            Number: The maximum value of the numpy array.
        """
        return self.apply(Max())

    def mean(self) -> Number:
        """Get the mean value of the numpy array.

        Returns:
            Number: The mean value of the numpy array.
        """
        return self.apply(Mean())

    def std(self) -> Number:
        """Get the standard deviation of the numpy array.

        Returns:
            Number: The standard deviation of the numpy array.
        """
        return self.apply(Std())

    @classmethod
    def validate_data(cls, data: np.ndarray) -> np.ndarray:
        """Validate the data to ensure it is a numpy array.

        Args:
            data (numpyp.ndarray): The numpy array to validate.
        
        Raises:
            TypeError: If the input data is not a numpy array.

        Returns:
            numpy.ndarray: The validated numpy array.
        """
        return cls._validate_dtype(data, (np.ndarray,))



import os, shutil
import jijmodeling as jm
from matplotlib import axes, figure

import numpy as np
import pandas as pd
import pytest

import jijbench as jb

from jijbench.visualization.metrics.plot import _get_violations_dict, MetricsPlot


@pytest.fixture(scope="function", autouse=True)
def pre_post_process():
    # preprocess
    yield
    # postprocess
    norm_path = os.path.normcase("./.jb_results")
    if os.path.exists(norm_path):
        shutil.rmtree(norm_path)


def solve():
    d = jm.Placeholder("d", dim=2)
    x = jm.Binary("x", shape=d.shape)
    i = jm.Element("i", d.shape[0])
    j = jm.Element("j", d.shape[1])
    problem = jm.Problem("simple_problem")
    problem += jm.Sum([i, j], d[i, j] * x[i, j])
    problem += jm.Constraint("onehot1", x[i, :] == 1, forall=i)
    problem += jm.Constraint("onehot2", x[:, j] == 1, forall=j)
    jm_sampleset_dict = {
        "record": {
            "solution": {
                "x": [
                    (([0, 1], [0, 1]), [1, 1], (2, 2)),
                    (([0, 1], [1, 0]), [1, 1], (2, 2)),
                    (([], []), [], (2, 2)),
                    (([0, 1], [0, 0]), [1, 1], (2, 2)),
                ]
            },
            "num_occurrences": [4, 3, 2, 1],
        },
        "evaluation": {
            "energy": [3.0, 24.0, 0.0, 20.0],
            "objective": [3.0, 24.0, 0.0, 17.0],
            "constraint_violations": {
                "onehot1": [0.0, 0.0, 2.0, 0.0],
                "onehot2": [0.0, 0.0, 2.0, 2.0],
            },
            "penalty": {},
        },
        "measuring_time": {"solve": None, "system": None, "total": None},
    }
    jm_sampleset = jm.SampleSet.from_serializable(jm_sampleset_dict)
    solving_time = jm.SolvingTime(
        **{"preprocess": 1.0, "solve": 1.0, "postprocess": 1.0}
    )
    jm_sampleset.measuring_time.solve = solving_time
    return jm_sampleset


def simple_func_for_boxplot(x: pd.Series) -> dict:
    return {"data1": np.array([1, 2, 3, 4]), "data2": np.array([1, 2, 3, 4])}


def test_metrics_plot_get_violations_dict():
    series = pd.Series(
        data=[np.array([1, 1]), np.array([0, 2])],
        index=["num_occurrences", "onehot_violations"],
    )
    violations_dict = _get_violations_dict(series)

    assert len(violations_dict.keys()) == 1
    assert (violations_dict["onehot_violations"] == np.array([0, 2])).all()


def test_metrics_plot_boxplot_return_value():
    num_multipliers = 2
    params = {"multipliers": [{} for _ in range(num_multipliers)]}
    solver_list = [solve]

    expect_num_fig = len(params["multipliers"]) * len(solver_list)

    bench = jb.Benchmark(
        params=params,
        solver=solver_list,
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig_ax_tuple = mplot.boxplot(f=simple_func_for_boxplot)
    assert len(fig_ax_tuple) == expect_num_fig
    assert type(fig_ax_tuple[0][0]) == figure.Figure
    assert type(fig_ax_tuple[0][1]) == axes.Subplot


def test_metrics_plot_boxplot_call_maplotlib_boxplot(mocker):
    m = mocker.patch("matplotlib.axes.Subplot.boxplot")

    num_multipliers = 2
    params = {"multipliers": [{} for _ in range(num_multipliers)]}
    solver_list = [solve]

    expect_num_call = len(params["multipliers"]) * len(solver_list)

    bench = jb.Benchmark(
        params=params,
        solver=solver_list,
    )
    result = bench()
    mplot = MetricsPlot(result)
    mplot.boxplot(f=simple_func_for_boxplot)

    assert m.call_count == expect_num_call


def test_metrics_plot_boxplot_arg_figsize():
    figwidth, figheight = 8, 4

    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig_ax_tuple = mplot.boxplot(
        f=simple_func_for_boxplot,
        figsize=(figwidth, figheight),
    )
    fig, ax = fig_ax_tuple[0]

    assert fig.get_figwidth() == 8
    assert fig.get_figheight() == 4


def test_metrics_plot_boxplot_arg_title(mocker):
    m = mocker.patch("matplotlib.figure.Figure.suptitle")

    title = ["title1", "title2"]

    num_multipliers = 2
    params = {"multipliers": [{} for _ in range(num_multipliers)]}
    solver_list = [solve]

    bench = jb.Benchmark(
        params=params,
        solver=solver_list,
    )
    result = bench()
    mplot = MetricsPlot(result)
    mplot.boxplot(
        f=simple_func_for_boxplot,
        title=title,
    )
    m.assert_has_calls(
        [
            mocker.call("title1", fontsize=None),
            mocker.call("title2", fontsize=None),
        ]
    )


def test_metrics_plot_boxplot_arg_title_fontsize(mocker):
    m = mocker.patch("matplotlib.figure.Figure.suptitle")

    title = ["title1", "title2"]
    title_fontsize = 12

    num_multipliers = 2
    params = {"multipliers": [{} for _ in range(num_multipliers)]}
    solver_list = [solve]

    bench = jb.Benchmark(
        params=params,
        solver=solver_list,
    )
    result = bench()
    mplot = MetricsPlot(result)
    mplot.boxplot(
        f=simple_func_for_boxplot,
        title=title,
        title_fontsize=title_fontsize,
    )
    m.assert_has_calls(
        [
            mocker.call("title1", fontsize=title_fontsize),
            mocker.call("title2", fontsize=title_fontsize),
        ]
    )


def test_metrics_plot_boxplot_arg_xticklabels():
    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig_ax_tuple = mplot.boxplot(
        f=simple_func_for_boxplot,
    )
    ax = fig_ax_tuple[0][1]
    xticklabels = ax.get_xticklabels()

    assert xticklabels[0].get_text() == "data1"
    assert xticklabels[1].get_text() == "data2"


def test_metrics_plot_boxplot_arg_xticklabels_size(mocker):
    m = mocker.patch("matplotlib.axes.Subplot.set_xticklabels")
    xticklabels_size = 10
    data = {"data1": np.array([1, 2, 3, 4]), "data2": np.array([1, 2, 3, 4])}

    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    mplot.boxplot(
        f=simple_func_for_boxplot,
        xticklabels_size=xticklabels_size,
    )

    m.assert_called_once_with(data.keys(), size=10, rotation=None)


def test_metrics_plot_boxplot_arg_xticklabels_rotation(mocker):
    m = mocker.patch("matplotlib.axes.Subplot.set_xticklabels")
    xticklabels_rotation = 10
    data = {"data1": np.array([1, 2, 3, 4]), "data2": np.array([1, 2, 3, 4])}

    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    mplot.boxplot(
        f=simple_func_for_boxplot,
        xticklabels_rotation=xticklabels_rotation,
    )

    m.assert_called_once_with(data.keys(), size=None, rotation=10)


def test_metrics_plot_boxplot_arg_ylabel():
    ylabel = "ylabel"

    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig_ax_tuple = mplot.boxplot(
        f=simple_func_for_boxplot,
        ylabel=ylabel,
    )

    ax = fig_ax_tuple[0][1]
    assert ax.get_ylabel() == "ylabel"


def test_metrics_plot_boxplot_arg_ylabel_size(mocker):
    m = mocker.patch("matplotlib.axes.Subplot.set_ylabel")

    ylabel = "ylabel"
    ylabel_size = 10

    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    mplot.boxplot(f=simple_func_for_boxplot, ylabel=ylabel, ylabel_size=ylabel_size)

    m.assert_called_once_with("ylabel", size=ylabel_size)


def test_metrics_plot_boxplot_arg_yticks_default(mocker):
    m = mocker.patch("matplotlib.ticker.MaxNLocator.__init__", return_value=None)

    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    mplot.boxplot(f=simple_func_for_boxplot)

    m.assert_called_with(integer=True)


def test_metrics_plot_boxplot_arg_yticks():
    yticks = [0, 10]

    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig_ax_tuple = mplot.boxplot(
        f=simple_func_for_boxplot,
        yticks=yticks,
    )
    ax = fig_ax_tuple[0][1]
    assert (ax.get_yticks() == [0, 10]).all()


def test_metrics_plot_boxplot_arg_kwargs(mocker):
    m = mocker.patch("matplotlib.axes.Subplot.boxplot")

    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    mplot.boxplot(
        f=simple_func_for_boxplot,
        showmeans=True,
        meanline=True,
    )

    args, kwargs = m.call_args
    assert kwargs["showmeans"] is True
    assert kwargs["meanline"] is True


def test_metrics_plot_boxplot_violations_return_value():
    num_multipliers = 2
    params = {"multipliers": [{} for _ in range(num_multipliers)]}
    solver_list = [solve]

    expect_num_fig = len(params["multipliers"]) * len(solver_list)

    bench = jb.Benchmark(
        params=params,
        solver=solver_list,
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig_ax_tuple = mplot.boxplot_violations()
    assert len(fig_ax_tuple) == expect_num_fig
    assert type(fig_ax_tuple[0][0]) == figure.Figure
    assert type(fig_ax_tuple[0][1]) == axes.Subplot


def test_metrics_plot_boxplot_violations_call_maplotlib_boxplot(mocker):
    m = mocker.patch("matplotlib.axes.Subplot.boxplot")

    num_multipliers = 2
    params = {"multipliers": [{} for _ in range(num_multipliers)]}
    solver_list = [solve]

    expect_num_call = len(params["multipliers"]) * len(solver_list)

    bench = jb.Benchmark(
        params=params,
        solver=solver_list,
    )
    result = bench()
    mplot = MetricsPlot(result)
    mplot.boxplot_violations()

    assert m.call_count == expect_num_call


def test_metrics_plot_boxplot_violations_arg_figsize():
    figwidth, figheight = 8, 4

    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig_ax_tuple = mplot.boxplot_violations(
        figsize=(figwidth, figheight),
    )
    fig, ax = fig_ax_tuple[0]

    assert fig.get_figwidth() == 8
    assert fig.get_figheight() == 4


def test_metrics_plot_boxplot_violations_arg_title(mocker):
    m = mocker.patch("matplotlib.figure.Figure.suptitle")

    title = ["title1", "title2"]

    num_multipliers = 2
    params = {"multipliers": [{} for _ in range(num_multipliers)]}
    solver_list = [solve]

    bench = jb.Benchmark(
        params=params,
        solver=solver_list,
    )
    result = bench()
    mplot = MetricsPlot(result)
    mplot.boxplot_violations(
        title=title,
    )
    m.assert_has_calls(
        [
            mocker.call("title1", fontsize=None),
            mocker.call("title2", fontsize=None),
        ]
    )


def test_metrics_plot_boxplot_violations_arg_title_fontsize(mocker):
    m = mocker.patch("matplotlib.figure.Figure.suptitle")

    title = ["title1", "title2"]
    title_fontsize = 12

    num_multipliers = 2
    params = {"multipliers": [{} for _ in range(num_multipliers)]}
    solver_list = [solve]

    bench = jb.Benchmark(
        params=params,
        solver=solver_list,
    )
    result = bench()
    mplot = MetricsPlot(result)
    mplot.boxplot_violations(
        title=title,
        title_fontsize=title_fontsize,
    )
    m.assert_has_calls(
        [
            mocker.call("title1", fontsize=title_fontsize),
            mocker.call("title2", fontsize=title_fontsize),
        ]
    )


def test_metrics_plot_boxplot_violations_arg_xticklabels():
    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig_ax_tuple = mplot.boxplot_violations()
    ax = fig_ax_tuple[0][1]
    xticklabels = ax.get_xticklabels()

    assert xticklabels[0].get_text() == "onehot1_violations"
    assert xticklabels[1].get_text() == "onehot2_violations"


def test_metrics_plot_boxplot_violations_arg_xticklabels_size(mocker):
    m = mocker.patch("matplotlib.axes.Subplot.set_xticklabels")
    constraint_name_fontsize = 10

    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    mplot.boxplot_violations(
        constraint_name_fontsize=constraint_name_fontsize,
    )
    args, kwargs = m.call_args
    assert kwargs["size"] == 10


def test_metrics_plot_boxplot_violations_arg_xticklabels_rotation(mocker):
    m = mocker.patch("matplotlib.axes.Subplot.set_xticklabels")
    constraint_name_fontrotation = 10

    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    mplot.boxplot_violations(
        constraint_name_fontrotation=constraint_name_fontrotation,
    )
    args, kwargs = m.call_args
    assert kwargs["rotation"] == 10


def test_metrics_plot_boxplot_violations_arg_ylabel_default():
    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig_ax_tuple = mplot.boxplot_violations()

    ax = fig_ax_tuple[0][1]
    assert ax.get_ylabel() == "constraint violations"


def test_metrics_plot_boxplot_violations_arg_ylabel():
    ylabel = "ylabel"

    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig_ax_tuple = mplot.boxplot_violations(
        ylabel=ylabel,
    )

    ax = fig_ax_tuple[0][1]
    assert ax.get_ylabel() == "ylabel"


def test_metrics_plot_boxplot_violations_arg_ylabel_size(mocker):
    m = mocker.patch("matplotlib.axes.Subplot.set_ylabel")

    ylabel = "ylabel"
    ylabel_size = 10

    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    mplot.boxplot_violations(ylabel=ylabel, ylabel_size=ylabel_size)

    m.assert_called_once_with("ylabel", size=ylabel_size)


def test_metrics_plot_boxplot_violations_arg_yticks_default(mocker):
    m = mocker.patch("matplotlib.ticker.MaxNLocator.__init__", return_value=None)
    mocker.patch("matplotlib.axes.Subplot.axhline")

    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    mplot.boxplot_violations()

    m.assert_called_with(integer=True)


def test_metrics_plot_boxplot_violations_arg_yticks():
    yticks = [0, 10]

    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    fig_ax_tuple = mplot.boxplot_violations(
        yticks=yticks,
    )
    ax = fig_ax_tuple[0][1]
    assert (ax.get_yticks() == [0, 10]).all()


def test_metrics_plot_boxplot_violations_arg_kwargs_default(mocker):
    m = mocker.patch("matplotlib.axes.Subplot.boxplot")

    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    mplot.boxplot_violations()

    args, kwargs = m.call_args
    assert kwargs["showmeans"] is True
    assert kwargs["whis"] == [0, 100]


def test_metrics_plot_boxplot_violations_arg_kwargs(mocker):
    m = mocker.patch("matplotlib.axes.Subplot.boxplot")

    bench = jb.Benchmark(
        params={},
        solver=[solve],
    )
    result = bench()
    mplot = MetricsPlot(result)
    mplot.boxplot_violations(
        showmeans=True,
        meanline=True,
    )

    args, kwargs = m.call_args
    assert kwargs["showmeans"] is True
    assert kwargs["meanline"] is True


def test_metrics_plot_boxplot_violations_hline_indicate_no_violation(mocker):
    m = mocker.patch("matplotlib.axes.Subplot.axhline")

    num_multipliers = 2
    params = {"multipliers": [{} for _ in range(num_multipliers)]}
    solver_list = [solve]

    expect_num_call = len(params["multipliers"]) * len(solver_list)

    bench = jb.Benchmark(
        params=params,
        solver=solver_list,
    )
    result = bench()
    mplot = MetricsPlot(result)
    mplot.boxplot_violations()

    m.assert_has_calls(
        [
            mocker.call(0, xmin=0, xmax=1, color="gray", linestyle="dotted"),
            mocker.call(0, xmin=0, xmax=1, color="gray", linestyle="dotted"),
        ]
    )



from __future__ import annotations

from jijbench import datasets
from jijmodeling.type_annotations import PH_VALUES_INTERFACE


__all__ = []


def get_instance_data(
    problem_name, size="small"
) -> list[tuple[str, PH_VALUES_INTERFACE]]:
    cls = getattr(datasets, problem_name)()

    instance_data = []
    for name in cls.instance_names(size=size):
        instance_data.append((name, cls.get_instance(size=size, instance_name=name)))

    return instance_data



from __future__ import annotations

import abc
import pandas as pd
import typing as tp

from jijbench.node.base import FunctionNode
from jijbench.typing import DataNodeT, DataNodeT2

if tp.TYPE_CHECKING:
    from jijbench.containers.containers import Artifact, Record, Table


class Factory(FunctionNode[DataNodeT, DataNodeT2]):
    """An abstract base class for creating a new data node from a list of input nodes.

    Attributes:
        inputs (List[DataNodeT]): List of input data nodes.
        name (str or None): Name of the resulting data node.

    """

    @abc.abstractmethod
    def create(self, inputs: list[DataNodeT], name: str | None = None) -> DataNodeT2:
        """Abstract method to create a new data node.
        Subclasses must implement this method.

        Args:
            inputs (List[DataNodeT]): List of input data nodes.
            name (str or None): Name of the resulting data node.

        Returns:
            DataNodeT2: The resulting data node.

        """
        pass

    def operate(
        self, inputs: list[DataNodeT], name: str | None = None, **kwargs: tp.Any
    ) -> DataNodeT2:
        """Create a new data node from the given inputs.
        This method calls `create` method to create a new data node from the given inputs.

        Args:
            inputs (List[DataNodeT]): List of input data nodes.
            name (str or None, optional): Name of the resulting data node.
            **kwargs: Additional keyword arguments.

        Returns:
            DataNodeT2: The resulting data node.

        """
        return self.create(inputs, name, **kwargs)


class RecordFactory(Factory[DataNodeT, "Record"]):
    """A factory class for creating Record objects.

    This class creates Record objects from a list of input DataNode objects. It uses the `create` method to
    process the input DataNodes, extract their data and convert it into a pandas Series. The resulting Series is
    used to create the Record object. The class also includes a helper method `_to_nodes_from_sampleset` which
    is used to extract the relevant data from `jijmodeling.SampleSet` objects.
    """

    def create(
        self,
        inputs: list[DataNodeT],
        name: str = "",
    ) -> Record:
        """Create a Record object from the input DataNode objects.

        This method takes a list of input DataNode objects, processes them and converts them into a pandas
        Series. The resulting Series is used to create the Record object. If the input data is a `jijmodeling.SampleSet`
        and `is_parsed_sampleset` is set to True, the relevant data is extracted from the SampleSet using the
        `_to_nodes_from_sampleset` helper method.

        Args:
            inputs (list[DataNodeT]): A list of DataNode objects to be processed.
            name (str, optional): A name for the Record object. Defaults to "".
            is_parsed_sampleset (bool, optional): Whether to extract data from jijmodeling SampleSet objects. Defaults to True.

        Returns:
            Record: A Record object created from the processed input DataNode objects.
        """
        from jijbench.containers.containers import Record

        data = pd.Series({node.name: node for node in inputs})
        return Record(data, name)


class ArtifactFactory(Factory["Record", "Artifact"]):
    """A factory class for creating Artifact objects."""

    def create(self, inputs: list[Record], name: str = "") -> Artifact:
        """Creates an `Artifact` object using a list of `Record` inputs.

        Args:
            inputs (list[Record]): A list of `Record` objects to be used to create the `Artifact`.
            name (str, optional): Name of the `Artifact` object. Defaults to an empty string.

        Returns:
            Artifact: The created `Artifact` object.
        """
        from jijbench.containers.containers import Artifact

        data = {
            node.name
            if isinstance(node.name, tp.Hashable)
            else str(node.name): node.data.to_dict()
            for node in inputs
        }
        return Artifact(data, name)


class TableFactory(Factory["Record", "Table"]):
    """A factory class for creating Table objects."""

    def create(
        self,
        inputs: list[Record],
        name: str = "",
        index_name: str | None = None,
    ) -> Table:
        """Creates a `Table` object using a list of `Record` inputs.

        Args:
            inputs (list[Record]): A list of `Record` objects to be used to create the `Table`.
            name (str, optional): Name of the `Table` object. Defaults to an empty string.
            index_name (str | None, optional): Name of the index in the created `Table`. Defaults to None.

        Returns:
            Table: The created `Table` object.
        """
        from jijbench.containers.containers import Table

        data = pd.DataFrame({node.name: node.data for node in inputs}).T
        data.index.name = index_name
        return Table(data, name)



import jijmodeling as jm

import numpy as np
import pandas as pd
import pytest

import jijbench as jb

from jijbench.visualization.metrics.utils import (
    construct_experiment_from_samplesets,
    _create_fig_title_list,
    _df_has_valid_multipliers_column,
    _df_has_number_array_column_target_name,
    _df_has_number_column_target_name,
)


def test_construct_experiment_from_samplesets():
    num_occ1, num_occ2 = 1, 2

    sampleset1 = jm.SampleSet(
        record=jm.Record(
            solution={"x": [(([3],), [1], (6,))]},
            num_occurrences=[num_occ1],
        ),
        evaluation=jm.Evaluation(
            energy=[-8.0],
            objective=[-8.0],
            constraint_violations={"onehot": [0.0]},
            penalty={},
        ),
        measuring_time=jm.MeasuringTime(),
    )

    sampleset2 = jm.SampleSet(
        record=jm.Record(
            solution={"x": [(([3],), [1], (6,))]},
            num_occurrences=[num_occ2],
        ),
        evaluation=jm.Evaluation(
            energy=[-8.0],
            objective=[-8.0],
            constraint_violations={"onehot": [0.0]},
            penalty={},
        ),
        measuring_time=jm.MeasuringTime(),
    )

    samplesets = [sampleset1, sampleset2]

    experiment = construct_experiment_from_samplesets(samplesets)

    assert isinstance(experiment, jb.Experiment)
    assert len(experiment.table) == 2
    assert experiment.table["num_occurrences"].values[0] == np.array([num_occ1])
    assert experiment.table["num_occurrences"].values[1] == np.array([num_occ2])


def test_construct_experiment_from_samplesets_give_raw_sampleset():
    sampleset = jm.SampleSet(
        record=jm.Record(
            solution={"x": [(([3],), [1], (6,))]},
            num_occurrences=[1],
        ),
        evaluation=jm.Evaluation(
            energy=[-8.0],
            objective=[-8.0],
            constraint_violations={"onehot": [0.0]},
            penalty={},
        ),
        measuring_time=jm.MeasuringTime(),
    )

    experiment = construct_experiment_from_samplesets(sampleset)

    assert isinstance(experiment, jb.Experiment)
    assert len(experiment.table) == 1
    assert experiment.table["num_occurrences"].values[0] == np.array([1])


def test_construct_experiment_from_samplesets_additional_data():
    additional_data = {
        "data1": [0, 1],
        "data2": [np.array([2, 3]), np.array([4, 5])],
    }

    sampleset1 = jm.SampleSet(
        record=jm.Record(
            solution={"x": [(([3],), [1], (6,))]},
            num_occurrences=[1],
        ),
        evaluation=jm.Evaluation(
            energy=[-8.0],
            objective=[-8.0],
            constraint_violations={"onehot": [0.0]},
            penalty={},
        ),
        measuring_time=jm.MeasuringTime(),
    )

    sampleset2 = jm.SampleSet(
        record=jm.Record(
            solution={"x": [(([3],), [1], (6,))]},
            num_occurrences=[2],
        ),
        evaluation=jm.Evaluation(
            energy=[-8.0],
            objective=[-8.0],
            constraint_violations={"onehot": [0.0]},
            penalty={},
        ),
        measuring_time=jm.MeasuringTime(),
    )
    samplesets = [sampleset1, sampleset2]

    experiment = construct_experiment_from_samplesets(samplesets, additional_data)

    assert isinstance(experiment, jb.Experiment)
    assert len(experiment.table) == 2
    assert experiment.table["data1"].values[0] == 0
    assert experiment.table["data1"].values[1] == 1
    assert (experiment.table["data2"].values[0] == np.array([2, 3])).all()
    assert (experiment.table["data2"].values[1] == np.array([4, 5])).all()


def test_construct_experiment_from_samplesets_additional_data_invalid_length():
    additional_data = {
        "data1": [0],
    }

    sampleset1 = jm.SampleSet(
        record=jm.Record(
            solution={"x": [(([3],), [1], (6,))]},
            num_occurrences=[1],
        ),
        evaluation=jm.Evaluation(
            energy=[-8.0],
            objective=[-8.0],
            constraint_violations={"onehot": [0.0]},
            penalty={},
        ),
        measuring_time=jm.MeasuringTime(),
    )

    sampleset2 = jm.SampleSet(
        record=jm.Record(
            solution={"x": [(([3],), [1], (6,))]},
            num_occurrences=[2],
        ),
        evaluation=jm.Evaluation(
            energy=[-8.0],
            objective=[-8.0],
            constraint_violations={"onehot": [0.0]},
            penalty={},
        ),
        measuring_time=jm.MeasuringTime(),
    )
    samplesets = [sampleset1, sampleset2]

    with pytest.raises(TypeError):
        construct_experiment_from_samplesets(samplesets, additional_data)


params = {
    "input is title list": (["title1", "title2"], ["title1", "title2"]),
    "input is title string": ("title", ["title", "title"]),
    "input is None": (None, ["i: 1", "i: 2"]),
}


@pytest.mark.parametrize(
    "input_title, expect",
    list(params.values()),
    ids=list(params.keys()),
)
def test_create_fig_title_list(input_title, expect):
    series = pd.Series(
        data=[1, 2],
        index=["1", "2"],
    )
    series.index.names = ["i"]

    title_list = _create_fig_title_list(
        metrics=series,
        title=input_title,
    )
    assert title_list == expect


def test_create_fig_title_list_for_series_with_no_index():
    series = pd.Series(
        data=[1, 2],
        index=[None, None],
    )
    title_list = _create_fig_title_list(
        metrics=series,
        title=None,
    )
    assert title_list == ["", ""]


def test_create_fig_title_list_for_series_with_default_index():
    series = pd.Series(
        data=[1, 2],
    )
    title_list = _create_fig_title_list(
        metrics=series,
        title=None,
    )
    assert title_list == ["index: 0", "index: 1"]


def test_create_fig_title_list_for_invalid_input():
    invalid_input_title = 0

    series = pd.Series(
        data=[1, 2],
        index=["1", "2"],
    )
    series.index.names = ["i"]
    with pytest.raises(TypeError):
        _create_fig_title_list(
            metrics=series,
            title=invalid_input_title,
        )


params = {
    "no_multipliers_columns": ([[0, 1], [2, 3]], ["col_0", "col_1"], False),
    "first_multipliers_isnot_dict": ([[0]], ["multipliers"], False),
    "second_multipliers_isnot_dict": ([[{"onehot": 1}], [0]], ["multipliers"], False),
    "multipliers_key_isnot_str": ([[{0: 1}]], ["multipliers"], False),
    "multipliers_value_isnot_number": ([[{"onehot": "1"}]], ["multipliers"], False),
    "constraint_name_is_different": (
        [[{"onehot1": 1}], [{"onehot2": 2}]],
        ["multipliers"],
        False,
    ),
    "valid_case": ([[{"onehot1": 1}], [{"onehot1": 2}]], ["multipliers"], True),
}


@pytest.mark.parametrize(
    "data, columns, expect",
    list(params.values()),
    ids=list(params.keys()),
)
def test_df_has_valid_multipliers_column(data, columns, expect):
    def create_df(data, columns):
        df = pd.DataFrame(columns=columns)
        for i, row in enumerate(data):
            for j, element in enumerate(row):
                df.at[i, columns[j]] = object
                df.at[i, columns[j]] = element
        return df

    df = create_df(data, columns)
    assert _df_has_valid_multipliers_column(df) == expect


params = {
    "no_target_columns": ("number_array", [[0, 1], [2, 3]], ["col_0", "col_1"], False),
    "target_column_isnot_array": ("number_array", [[0]], ["number_array"], False),
    "target_column_element_isnot_number": (
        "number_array",
        [[["a"]]],
        ["number_array"],
        False,
    ),
    "target_column_is_valid_list": (
        "number_array",
        [[[0, 1]], [[2, 3]]],
        ["number_array"],
        True,
    ),
    "target_column_is_valid_nparray": (
        "number_array",
        [[np.array([0, 1])], [np.array([2, 3])]],
        ["number_array"],
        True,
    ),
}


@pytest.mark.parametrize(
    "target_column, data, columns, expect",
    list(params.values()),
    ids=list(params.keys()),
)
def test_df_has_number_array_column_target_name(target_column, data, columns, expect):
    def create_df(data, columns):
        df = pd.DataFrame(columns=columns)
        for i, row in enumerate(data):
            for j, element in enumerate(row):
                df.at[i, columns[j]] = object
                df.at[i, columns[j]] = element
        return df

    df = create_df(data, columns)
    assert (
        _df_has_number_array_column_target_name(df, column_name=target_column) == expect
    )


params = {
    "no_target_columns": ("number", [[0, 1], [2, 3]], ["col_0", "col_1"], False),
    "target_column_is_list": ("number", [[[0]]], ["number"], False),
    "target_column_is_str": ("number", [["0"]], ["number"], False),
    "target_column_is_number": (
        "number",
        [[0], [1]],
        ["number"],
        True,
    ),
}


@pytest.mark.parametrize(
    "target_column, data, columns, expect",
    list(params.values()),
    ids=list(params.keys()),
)
def test_df_has_number_column_target_name(target_column, data, columns, expect):
    def create_df(data, columns):
        df = pd.DataFrame(columns=columns)
        for i, row in enumerate(data):
            for j, element in enumerate(row):
                df.at[i, columns[j]] = object
                df.at[i, columns[j]] = element
        return df

    df = create_df(data, columns)
    assert _df_has_number_column_target_name(df, column_name=target_column) == expect
